# It has been modified for vendored imports.
from synapse.vendor.cashaddress.crypto import *
from synapse.vendor.cashaddress.base58 import b58decode_check, b58encode_check
import sys


class InvalidAddress(Exception):
    pass


class Address:
    VERSION_MAP = {
        'legacy': [
            ('P2SH', 5, False),
            ('P2PKH', 0, False),
            ('P2SH-TESTNET', 196, True),
            ('P2PKH-TESTNET', 111, True)
        ],
        'cash': [
            ('P2SH', 8, False),
            ('P2PKH', 0, False),
            ('P2SH-TESTNET', 8, True),
            ('P2PKH-TESTNET', 0, True)
        ]
    }
    MAINNET_PREFIX = 'bitcoincash'
    TESTNET_PREFIX = 'bchtest'

    def __init__(self, version, payload, prefix=None):
        self.version = version
        self.payload = payload
        if prefix:
            self.prefix = prefix
        else:
            if Address._address_type('cash', self.version)[2]:
                self.prefix = self.TESTNET_PREFIX
            else:
                self.prefix = self.MAINNET_PREFIX

    def __str__(self):
        return 'version: {}\npayload: {}\nprefix: {}'.format(self.version, self.payload, self.prefix)

    def legacy_address(self):
        version_int = Address._address_type('legacy', self.version)[1]
        return b58encode_check(Address.code_list_to_string([version_int] + self.payload))

    def cash_address(self):
        version_int = Address._address_type('cash', self.version)[1]
        payload = [version_int] + self.payload
        payload = convertbits(payload, 8, 5)
        checksum = calculate_checksum(self.prefix, payload)
        return self.prefix + ':' + b32encode(payload + checksum)

    @staticmethod
    def code_list_to_string(code_list):
        if sys.version_info > (3, 0):
            output = bytes()
            for code in code_list:
                output += bytes([code])
        else:
            output = ''
            for code in code_list:
                output += chr(code)
        return output

    @staticmethod
    def _address_type(address_type, version):
        for mapping in Address.VERSION_MAP[address_type]:
            if mapping[0] == version or mapping[1] == version:
                return mapping
        raise InvalidAddress('Could not determine address version')

    @staticmethod
    def from_string(address_string):
        try:
            address_string = str(address_string)
        except Exception:
            raise InvalidAddress('Expected string as input')
        if ':' not in address_string:
            return Address._legacy_string(address_string)
        else:
            return Address._cash_string(address_string)

    @staticmethod
    def _legacy_string(address_string):
        try:
            decoded = bytearray(b58decode_check(address_string))
        except ValueError:
            raise InvalidAddress('Could not decode legacy address')
        version = Address._address_type('legacy', decoded[0])[0]
        payload = list()
        for letter in decoded[1:]:
            payload.append(letter)
        return Address(version, payload)

    @staticmethod
    def _cash_string(address_string):
        if address_string.upper() != address_string and address_string.lower() != address_string:
            raise InvalidAddress('Cash address contains uppercase and lowercase characters')
        address_string = address_string.lower()
        colon_count = address_string.count(':')
        if colon_count == 0:
            address_string = Address.MAINNET_PREFIX + ':' + address_string
        elif colon_count > 1:
            raise InvalidAddress('Cash address contains more than one colon character')
        prefix, base32string = address_string.split(':')
        decoded = b32decode(base32string)
        if not verify_checksum(prefix, decoded):
            raise InvalidAddress('Bad cash address checksum')
        converted = convertbits(decoded, 5, 8)
        version = Address._address_type('cash', converted[0])[0]
        if prefix == Address.TESTNET_PREFIX:
            version += '-TESTNET'
        payload = converted[1:-6]
        return Address(version, payload, prefix)


def to_cash_address(address):
    return Address.from_string(address).cash_address()


def to_legacy_address(address):
    return Address.from_string(address).legacy_address()


def is_valid(address):
    try:
        Address.from_string(address)
        return True
    except InvalidAddress:
        return False



'''
Utilities for handling data files embedded within python packages.
'''
import os

import synapse.lib.dyndeps as s_dyndeps

def openDatFile(datpath):
    '''
    Open a file-like object using a pkg relative path.

    Example:

        fd = openDatFile('foopkg.barpkg/wootwoot.bin')
    '''
    pkgname, filename = datpath.split('/', 1)

    pkgmod = s_dyndeps.getDynMod(pkgname)

    # are we a regular file?
    pkgfile = os.path.abspath(pkgmod.__file__)
    if os.path.isfile(pkgfile):
        dirname = os.path.dirname(pkgfile)
        datname = os.path.join(dirname, filename)
        return open(datname, 'rb')



#!/usr/bin/env python
import os
import sys
import subprocess

from setuptools import setup, find_packages
from setuptools.command.install import install

VERSION = '2.88.0'

class VerifyVersionCommand(install):
    """Custom command to verify that the git tag matches our version"""
    description = 'verify that the git tag matches our version'

    def run(self):
        tag = os.getenv('CIRCLE_TAG', '')
        tag = tag.lstrip('v')

        if tag != VERSION:
            info = f"Git tag: {tag} does not match the version of this app: {VERSION}"
            sys.exit(info)

class ReplaceCommitVersion(install):
    description = 'Replace the embedded commit information with our current git commit'
    def run(self):
        try:
            ret = subprocess.run(['git', 'rev-parse', 'HEAD'],
                                 capture_output=True,
                                 timeout=15,
                                 check=False,
                                 text=True,
                                 )
        except Exception as e:
            print(f'Error grabbing commit: {e}')
            return 1
        else:
            commit = ret.stdout.strip()
        fp = './synapse/lib/version.py'
        with open(fp, 'rb') as fd:
            buf = fd.read()
        content = buf.decode()
        new_content = content.replace("commit = ''", f"commit = '{commit}'")
        if content == new_content:
            print(f'Unable to insert commit into {fp}')
            return 1
        with open(fp, 'wb') as fd:
            _ = fd.write(new_content.encode())
        print(f'Inserted commit {commit} into {fp}')
        return 0


long_description_content_type = None
long_description = None
readme = './README.rst'
if os.path.exists(readme):

    print(f'Adding {readme} contents as the long description.')

    with open('./README.rst', 'rb') as fd:
        buf = fd.read()
    long_description = buf.decode()
    long_description_content_type = 'text/x-rst'

setup(
    name='synapse',
    version=VERSION,
    description='Synapse Intelligence Analysis Framework',
    long_description=long_description,
    long_description_content_type=long_description_content_type,
    author='The Vertex Project LLC',
    author_email='synapse@vertex.link',
    url='https://github.com/vertexproject/synapse',
    license='Apache License 2.0',

    packages=find_packages(exclude=['scripts',
                                    ]),

    include_package_data=True,

    install_requires=[
        'pyOpenSSL>=21.0.0,<22.0.0',
        'cryptography>=36.0.0,<39.0.0',
        'msgpack>=1.0.2,<1.1.0',
        'xxhash>=1.4.4,<3.1.0',
        'lmdb>=1.2.1,<1.4.0',
        'tornado>=6.1.0,<7.0.0',
        'regex>=2021.7.1',
        'PyYAML>=5.4,<6.1.0',
        'aiohttp>=3.8.1,<4.0',
        'aiohttp-socks>=0.6.1,<0.7.0',
        'aioimaplib>=0.9.0,<0.10.0',
        'aiosmtplib>=1.1.6,<2.0',
        'prompt-toolkit>=3.0.4,<3.1.0',
        'lark-parser==0.11.2',
        'Pygments>=2.7.4,<2.11.3',
        'packaging>=20.0,<22.0',
        'fastjsonschema>=2.15.3,<2.16',
        'stix2-validator>=3.0.0,<4.0.0',
        'vcrpy>=4.1.1,<4.2.0',
        'base58>=2.1.0,<2.2.0',
        'python-bitcoinlib==0.11.0',
        'pycryptodome>=3.11.0,<3.15.0',
        'typing-extensions>=3.7.4,<5.0.0',  # synapse.vendor.xrpl req
        'scalecodec>=1.0.2,<1.0.35',  # synapse.vendor.substrateinterface req
        'cbor2>=5.4.1,<5.4.3',
        'bech32==1.2.0',
        'oauthlib>=3.1.1,<4.0.0',
        'idna==3.3',
        'python-dateutil>=2.8,<3.0',
        'pytz>=2021.3,<2022',
    ],

    extras_require={
        'docs': [
            'nbconvert==5.6.1',
            'jupyter-client<=6.1.12',
            'jupyter>=1.0.0,<2.0.0',
            'hide-code>=0.5.2,<0.5.3',
            'nbstripout>=0.3.3,<1.0.0',
            'sphinx>=4.2.0,<5.0.0',
            'sphinx-rtd-theme>=1.0.0,<2.0.0',
            'sphinx-notfound-page==0.8',
        ],
        'dev': [
            'pytest>=6.2.5,<7.0.0',
            'autopep8>=1.5.4,<2.0.0',
            'pytest-cov>=3.0.0,<4.0.0',
            'pycodestyle>=2.8.0,<3.0.0',
            'bump2version>=1.0.1,<1.1.0',
            'pytest-xdist>=2.4.0,<3.0.0',
            'coverage>=6.2,<6.3',
        ],
    },

    classifiers=[
        'Development Status :: 5 - Production/Stable',

        'License :: OSI Approved :: Apache Software License',

        'Topic :: System :: Clustering',
        'Topic :: System :: Distributed Computing',
        'Topic :: System :: Software Distribution',

        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',
    ],
    cmdclass={
        'verify': VerifyVersionCommand,
        'setcommit': ReplaceCommitVersion,
    },
)



import io
import json
import logging

import synapse.common as s_common
import synapse.lib.structlog as s_structlog

import synapse.tests.utils as s_test
import synapse.exc as s_exc
logger = logging.getLogger(__name__)


class ZDE(s_exc.SynErr): pass


class StructLogTest(s_test.SynTest):

    def test_structlog(self):
        stream = io.StringIO()
        handler = logging.StreamHandler(stream=stream)
        formatter = s_structlog.JsonFormatter()
        handler.setFormatter(formatter)
        logger.addHandler(handler)

        logger.warning('Test message 1')
        logger.error('Test message 2')
        iden = s_common.guid()
        logger.error('Extra test', extra={'synapse': {'foo': 'bar', 'iden': iden, 'time': 0}})

        def foo():
            _ = 1 / 0
        def bar():
            try:
                foo()
            except ZeroDivisionError as e:
                raise ZDE(mesg='ZDE test', args=(1, 0), buffer='vertex'.encode()) from e
        try:
            bar()
        except s_exc.SynErr:
            logger.exception('Exception handling')

        logger.warning('Unicode is cool for 程序员!')

        data = stream.getvalue()

        # There is a trailing \n on the stream
        raw_mesgs = [m for m in data.split('\n') if m]
        mesgs = [json.loads(m) for m in raw_mesgs]
        self.len(5, mesgs)

        mesg = mesgs[0]
        self.eq(set(mesg.keys()), {'message', 'logger', 'level', 'time'})
        lnfo = mesg.get('logger')
        self.eq(set(lnfo.keys()), {'name', 'process', 'filename', 'func'})
        self.eq(mesg.get('message'), 'Test message 1')
        self.eq(mesg.get('level'), 'WARNING')

        mesg = mesgs[1]
        self.eq(mesg.get('message'), 'Test message 2')
        self.eq(mesg.get('level'), 'ERROR')

        mesg = mesgs[2]
        self.eq(mesg.get('message'), 'Extra test')
        self.eq(mesg.get('level'), 'ERROR')
        self.eq(mesg.get('foo'), 'bar')
        self.eq(mesg.get('iden'), iden)
        self.ne(mesg.get('time'), 0)  # time was not overwritten by the extra

        mesg = mesgs[3]
        self.eq(mesg.get('message'), 'Exception handling')
        self.eq(mesg.get('level'), 'ERROR')
        erfo = mesg.get('err')

        etb = erfo.get('etb')
        self.isin('Traceback', etb)
        self.isin('_ = 1 / 0', etb)
        self.isin('The above exception was the direct cause of the following exception:', etb)
        self.isin('ZeroDivisionError: division by zero', etb)
        self.isin("""test_lib_structlog.ZDE: ZDE: args=(1, 0) buffer=b'vertex' mesg='ZDE test'""", etb)
        self.eq(erfo.get('errname'), 'ZDE')
        self.eq(erfo.get('mesg'), 'ZDE test')
        self.eq(erfo.get('args'), (1, 0))
        self.eq(erfo.get('buffer'), "b'vertex'")

        mesg = mesgs[4]
        rawm = raw_mesgs[4]
        self.isin(r'Unicode is cool for \u7a0b\u5e8f\u5458!', rawm)
        self.eq(mesg.get('message'), 'Unicode is cool for 程序员!')



import os

import synapse.tests.utils as s_test

import synapse.lib.spooled as s_spooled

class SpooledTest(s_test.SynTest):

    async def test_spooled_set(self):
        async with await s_spooled.Set.anit(size=2) as sset:
            self.len(0, sset)

            await sset.add(10)
            self.true(10 in sset)
            self.len(1, sset)

            await sset.add(10)
            self.len(1, sset)

            sset.discard(10)
            self.false(10 in sset)

            self.len(0, sset)
            self.false(sset.fallback)

            await sset.add(10)

            # Trigger fallback
            await sset.add(20)
            await sset.add(30)
            await sset.add(None)

            self.len(4, sset)

            await sset.add(20)
            self.len(4, sset)

            self.nn(sset.slab)
            self.true(sset.fallback)

            self.true(10 in sset)
            sset.discard(10)
            self.false(10 in sset)
            self.len(3, sset)

            sset.discard(10)
            self.len(3, sset)

            self.true(None in sset)

            self.true(os.path.isdir(sset.slab.path))

        self.false(os.path.isdir(sset.slab.path))
        self.false(os.path.isfile(sset.slab.optspath))

    async def test_spooled_set_dir(self):
        with self.getTestDir() as dirn:
            async with await s_spooled.Set.anit(dirn=dirn, size=2) as sset:
                await sset.add(10)
                await sset.add(20)
                await sset.add(30)
                self.true(os.path.isdir(sset.slab.path))
                self.true(os.path.abspath(sset.slab.path).startswith(dirn))

    async def test_spooled_dict(self):

        async def runtest(x):
            await x.set(10, 'hehe')
            self.eq(x.get(10), 'hehe')
            self.eq(x.get(20, 'newp'), 'newp')
            await x.set(20, 'haha')
            await x.set(30, 'hoho')
            self.eq(x.get(20), 'haha')
            self.eq(x.get(40, 'newp'), 'newp')
            self.len(3, x)
            self.eq('hehe', x.get(10))
            self.eq(list(x.items()), ((10, 'hehe'), (20, 'haha'), (30, 'hoho')))
            self.eq(list(x.keys()), (10, 20, 30))
            self.true(x.has(20))
            self.false(x.has(99))

        async with await s_spooled.Dict.anit(size=2) as sd0:
            await runtest(sd0)

        async with await s_spooled.Dict.anit(size=1000) as sd1:
            await runtest(sd1)



import os
import sys
import base64
import asyncio
import logging
import argparse

import regex

import synapse.exc as s_exc
import synapse.common as s_common
import synapse.telepath as s_telepath

import synapse.lib.output as s_output
import synapse.lib.dyndeps as s_dyndeps

logger = logging.getLogger(__name__)

wflownamere = regex.compile(r'^([\w-]+)\.yaml$')

def chopSemVer(vers):
    return tuple([int(x) for x in vers.split('.')])

def getStormStr(fn):
    if not os.path.isfile(fn):
        raise s_exc.NoSuchFile(mesg='Storm file {} not found'.format(fn), path=fn)

    with open(fn, 'rb') as f:
        return f.read().decode()

def loadOpticFiles(pkgdef, path):

    pkgfiles = pkgdef['optic']['files']

    abspath = s_common.genpath(path)
    for root, dirs, files, in os.walk(path):

        for name in files:

            if name.startswith('.'):  # pragma: no cover
                continue

            fullname = s_common.genpath(root, name)
            if not os.path.isfile(fullname):  # pragma: no cover
                continue

            pkgfname = fullname[len(abspath) + 1:]

            with open(fullname, 'rb') as fd:
                pkgfiles[pkgfname] = {
                    'file': base64.b64encode(fd.read()).decode(),
                }

def loadOpticWorkflows(pkgdef, path):

    wdefs = pkgdef['optic']['workflows']

    for root, dirs, files in os.walk(path):

        for name in files:

            match = wflownamere.match(name)

            if match is None:
                logger.warning('Skipping workflow "%s" that does not match pattern "%s"' % (name, wflownamere.pattern))
                continue

            wname = match.groups()[0]

            fullname = s_common.genpath(root, name)
            if not os.path.isfile(fullname):  # pragma: no cover
                continue

            wdefs[wname] = s_common.yamlload(fullname)

def tryLoadPkgProto(fp, opticdir=None, readonly=False):
    '''
    Try to get a Storm Package prototype from disk with or without inline documentation.

    Args:
        fp (str): Path to the package .yaml file on disk.
        opticdir (str): Path to optional Optic module code to add to the Storm Package.
        readonly (bool): If set, open files in read-only mode. If files are missing, that will raise a NoSuchFile
                         exception.

    Returns:
        dict: A Storm package definition.
    '''
    try:
        return loadPkgProto(fp, opticdir=opticdir, readonly=readonly)
    except s_exc.NoSuchFile:
        return loadPkgProto(fp, opticdir=opticdir, no_docs=True, readonly=readonly)

def loadPkgProto(path, opticdir=None, no_docs=False, readonly=False):
    '''
    Get a Storm Package definition from disk.

    Args:
        fp (str): Path to the package .yaml file on disk.
        opticdir (str): Path to optional Optic module code to add to the Storm Package.
        no_docs (bool): If true, omit inline documentation content if it is not present on disk.
        readonly (bool): If set, open files in read-only mode. If files are missing, that will raise a NoSuchFile
                         exception.

    Returns:
        dict: A Storm package definition.
    '''

    full = s_common.genpath(path)
    pkgdef = s_common.yamlload(full)

    if isinstance(pkgdef['version'], str):
        pkgdef['version'] = chopSemVer(pkgdef['version'])

    protodir = os.path.dirname(full)
    pkgname = pkgdef.get('name')

    genopts = pkgdef.pop('genopts', {})

    logodef = pkgdef.get('logo')
    if logodef is not None:

        path = logodef.pop('path', None)

        if path is not None:
            with s_common.reqfile(protodir, path) as fd:
                logodef['file'] = base64.b64encode(fd.read()).decode()

        if logodef.get('mime') is None:
            mesg = 'Mime type must be specified for logo file.'
            raise s_exc.BadPkgDef(mesg=mesg)

        if logodef.get('file') is None:
            mesg = 'Logo def must contain path or file.'
            raise s_exc.BadPkgDef(mesg=mesg)

    for docdef in pkgdef.get('docs', ()):

        if docdef.get('title') is None:
            mesg = 'Each entry in docs must have a title.'
            raise s_exc.BadPkgDef(mesg=mesg)

        if no_docs:
            docdef['content'] = ''
            continue

        path = docdef.pop('path', None)
        if path is not None:
            with s_common.reqfile(protodir, path) as fd:
                docdef['content'] = fd.read().decode()

        if docdef.get('content') is None:
            mesg = 'Docs entry has no path or content.'
            raise s_exc.BadPkgDef(mesg=mesg)

    for mod in pkgdef.get('modules', ()):

        name = mod.get('name')

        basename = name
        if genopts.get('dotstorm', False):
            basename = f'{basename}.storm'

        mod_path = s_common.genpath(protodir, 'storm', 'modules', basename)
        if readonly:
            mod['storm'] = getStormStr(mod_path)
        else:
            with s_common.genfile(mod_path) as fd:
                mod['storm'] = fd.read().decode()

    for extmod in pkgdef.get('external_modules', ()):
        fpth = extmod.get('file_path')
        if fpth is not None:
            extmod['storm'] = getStormStr(fpth)
        else:
            path = extmod.get('package_path')
            extpkg = s_dyndeps.tryDynMod(extmod.get('package'))
            extmod['storm'] = extpkg.getAssetStr(path)

        extname = extmod.get('name')
        extmod['name'] = f'{pkgname}.{extname}'

        pkgdef.setdefault('modules', [])
        pkgdef['modules'].append(extmod)

    pkgdef.pop('external_modules', None)

    for cmd in pkgdef.get('commands', ()):
        name = cmd.get('name')

        basename = name
        if genopts.get('dotstorm'):
            basename = f'{basename}.storm'

        cmd_path = s_common.genpath(protodir, 'storm', 'commands', basename)
        if readonly:
            cmd['storm'] = getStormStr(cmd_path)
        else:
            with s_common.genfile(cmd_path) as fd:
                cmd['storm'] = fd.read().decode()

    wflowdir = s_common.genpath(protodir, 'workflows')
    if os.path.isdir(wflowdir):
        pkgdef.setdefault('optic', {})
        pkgdef['optic'].setdefault('workflows', {})
        loadOpticWorkflows(pkgdef, wflowdir)

    if opticdir is None:
        opticdir = s_common.genpath(protodir, 'optic')

    if os.path.isdir(opticdir):
        pkgdef.setdefault('optic', {})
        pkgdef['optic'].setdefault('files', {})
        loadOpticFiles(pkgdef, opticdir)

    return pkgdef


prog = 'synapse.tools.genpkg'
desc = 'A tool for generating/pushing storm packages from YAML prototypes.'

async def main(argv, outp=s_output.stdout):

    pars = argparse.ArgumentParser()
    pars.add_argument('--push', metavar='<url>', help='A telepath URL of a Cortex or PkgRepo.')
    pars.add_argument('--save', metavar='<path>', help='Save the completed package JSON to a file.')
    pars.add_argument('--optic', metavar='<path>', help='Load Optic module files from a directory.')
    pars.add_argument('--no-build', action='store_true',
                      help='Treat pkgfile argument as an already-built package')
    pars.add_argument('--no-docs', default=False, action='store_true',
                      help='Do not require docs to be present and replace any doc content with empty strings.')
    pars.add_argument('pkgfile', metavar='<pkgfile>',
                      help='Path to a storm package prototype .yaml file, or a completed package .json/.yaml file.')

    opts = pars.parse_args(argv)

    if opts.no_build:
        pkgdef = s_common.yamlload(opts.pkgfile)
        if not pkgdef:
            outp.printf(f'Unable to load pkgdef from [{opts.pkgfile}]')
            return 1
        if opts.save:
            outp.printf(f'File {opts.pkgfile} is treated as already built (--no-build); incompatible with --save.')
            return 1
    else:
        pkgdef = loadPkgProto(opts.pkgfile, opticdir=opts.optic, no_docs=opts.no_docs)

    if not opts.save and not opts.push:
        outp.printf('Neither --push nor --save provided.  Nothing to do.')
        return 1

    if opts.save:
        s_common.jssave(pkgdef, opts.save)

    if opts.push:

        path = s_common.genpath('~/.syn/telepath.yaml')
        fini = await s_telepath.loadTeleEnv(path)

        async with await s_telepath.openurl(opts.push) as core:
            await core.addStormPkg(pkgdef)

        if fini is not None:  # pragma: no cover
            await fini()

    return 0

if __name__ == '__main__':  # pragma: no cover
    sys.exit(asyncio.run(main(sys.argv[1:])))



import copy
import json
import asyncio
import logging

import synapse.exc as s_exc

import synapse.lib.coro as s_coro
import synapse.lib.config as s_config
import synapse.lib.stormtypes as s_stormtypes

logger = logging.getLogger(__name__)

def runJsSchema(schema, item, use_default=True):
    # This is a target function for s_coro.spawn
    func = s_config.getJsValidator(schema, use_default=use_default)
    resp = func(item)
    return resp

def compileJsSchema(schema, use_default=True):
    # This is a target function for s_coro.spawn
    _ = s_config.getJsValidator(schema, use_default=use_default)
    return True

@s_stormtypes.registry.registerType
class JsonSchema(s_stormtypes.StormType):
    '''
    A JsonSchema validation object for use in validating data structures in Storm.
    '''
    _storm_typename = 'storm:json:schema'
    _storm_locals = (
        {'name': 'schema',
         'desc': 'The schema belonging to this object.',
         'type': {'type': 'function', '_funcname': '_schema',
                  'returns': {'type': 'dict', 'desc': 'A copy of the schema used for this object.', }}},
        {'name': 'validate',
         'desc': 'Validate a structure against the Json Schema',
         'type': {'type': 'function', '_funcname': '_validate',
                  'args': ({'name': 'item', 'type': 'prim',
                            'desc': 'A JSON structure to validate (dict, list, etc...)', },
                  ),
                  'returns': {'type': 'list',
                              'desc': 'An ($ok, $valu) tuple. If $ok is True, then $valu should be used as the '
                                      'validated data structure. If $ok is False, $valu is a dictionary with a "mesg" '
                                      'key.'}}},
    )
    _ismutable = False

    def __init__(self, runt, schema, use_default=True):
        s_stormtypes.StormType.__init__(self, None)
        self.runt = runt
        self.schema = schema
        self.use_default = use_default
        self.locls.update(self.getObjLocals())

    async def stormrepr(self):
        return f'{self._storm_typename}: {self.schema}'

    def getObjLocals(self):
        return {
            'schema': self._schema,
            'validate': self._validate,
        }

    async def _schema(self):
        return copy.deepcopy(self.schema)

    async def _validate(self, item):
        item = await s_stormtypes.toprim(item)

        try:
            result = await s_coro.spawn((runJsSchema, (self.schema, item), {'use_default': self.use_default}))
        except s_exc.SchemaViolation as e:
            return False, {'mesg': e.get('mesg')}
        else:
            return True, result

@s_stormtypes.registry.registerLib
class JsonLib(s_stormtypes.Lib):
    '''
    A Storm Library for interacting with Json data.
    '''
    _storm_locals = (
        {'name': 'load', 'desc': 'Parse a JSON string and return the deserialized data.',
         'type': {'type': 'function', '_funcname': '_jsonLoad',
                  'args': (
                      {'name': 'text', 'type': 'str', 'desc': 'The string to be deserialized.', },
                  ),
                  'returns': {'type': 'prim', 'desc': 'The JSON deserialized object.', }}},
        {'name': 'save', 'desc': 'Save an object as a JSON string.',
         'type': {'type': 'function', '_funcname': '_jsonSave',
                  'args': (
                      {'name': 'item', 'type': 'any', 'desc': 'The item to be serialized as a JSON string.', },
                  ),
                  'returns': {'type': 'str', 'desc': 'The JSON serialized object.', }}},
        {'name': 'schema', 'desc': 'Get a JS schema validation object.',
         'type': {'type': 'function', '_funcname': '_jsonSchema',
                  'args': (
                      {'name': 'schema', 'type': 'dict', 'desc': 'The JsonSchema to use.'},
                      {'name': 'use_default', 'type': 'boolean', 'default': True,
                       'desc': 'Whether to insert default schema values into the validated data structure.'},
                  ),
                  'returns': {'type': 'storm:json:schema',
                              'desc': 'A validation object that can be used to validate data structures.'}}},
    )
    _storm_lib_path = ('json',)

    def getObjLocals(self):
        return {
            'save': self._jsonSave,
            'load': self._jsonLoad,
            'schema': self._jsonSchema,
        }

    async def _jsonSave(self, item):
        try:
            item = await s_stormtypes.toprim(item)
            return json.dumps(item)
        except Exception as e:
            mesg = f'Argument is not JSON compatible: {item}'
            raise s_exc.MustBeJsonSafe(mesg=mesg)

    async def _jsonLoad(self, text):
        text = await s_stormtypes.tostr(text)
        try:
            return json.loads(text, strict=True)
        except Exception as e:
            mesg = f'Text is not valid JSON: {text}'
            raise s_exc.BadJsonText(mesg=mesg)

    async def _jsonSchema(self, schema, use_default=True):
        schema = await s_stormtypes.toprim(schema)
        use_default = await s_stormtypes.tobool(use_default)
        # We have to ensure that we have a valid schema for making the object.
        try:
            await s_coro.spawn((compileJsSchema, (schema,), {'use_default': use_default}))
        except asyncio.CancelledError:  # pragma: no cover
            raise
        except Exception as e:
            raise s_exc.StormRuntimeError(mesg=f'Unable to compile Json Schema: {str(e)}', schema=schema) from e
        return JsonSchema(self.runt, schema, use_default=use_default)



import hashlib

class HashSet:

    def __init__(self):

        self.size = 0

        # BEWARE ORDER MATTERS FOR guid()
        self.hashes = (
            ('md5', hashlib.md5()),
            ('sha1', hashlib.sha1()),
            ('sha256', hashlib.sha256()),
            ('sha512', hashlib.sha512())
        )

    def guid(self):
        '''
        Use elements from this hash set to create a unique
        (re)identifier.
        '''
        iden = hashlib.md5()
        props = {'size': self.size}

        for name, item in self.hashes:
            iden.update(item.digest())
            props[name] = item.hexdigest()

        return iden.hexdigest(), props

    def eatfd(self, fd):
        '''
        Consume all the bytes from a file like object.

        Example:

            hset = HashSet()
            hset.eatfd(fd)

        '''
        fd.seek(0)
        byts = fd.read(10000000)
        while byts:
            self.update(byts)
            byts = fd.read(10000000)

        return self.guid()

    def update(self, byts):
        '''
        Update all the hashes in the set with the given bytes.
        '''
        self.size += len(byts)
        [h[1].update(byts) for h in self.hashes]

    def digests(self):
        '''
        Get a list of (name, bytes) tuples for the hashes in the hashset.
        '''
        return [(name, item.digest()) for (name, item) in self.hashes]



import asyncio
import hashlib
import datetime
from unittest import mock
from datetime import timezone as tz

import synapse.exc as s_exc
import synapse.common as s_common
import synapse.tests.utils as s_t_utils

import synapse.lib.boss as s_boss
import synapse.lib.hive as s_hive
import synapse.lib.nexus as s_nexus
import synapse.lib.lmdbslab as s_lmdbslab

import synapse.lib.agenda as s_agenda
from synapse.lib.agenda import TimeUnit as s_tu

class AgendaTest(s_t_utils.SynTest):
    def test_apptreq(self):

        # Invalid combinations
        self.raises(s_exc.BadTime, s_agenda.ApptRec, {s_tu.DAYOFWEEK: 3, s_tu.DAYOFMONTH: 4})
        self.raises(s_exc.BadTime, s_agenda.ApptRec, {s_tu.HOUR: 1}, s_tu.MINUTE)
        self.raises(s_exc.BadTime, s_agenda.ApptRec, {}, None)
        self.raises(s_exc.BadTime, s_agenda.ApptRec, {}, s_tu.DAYOFMONTH)
        self.raises(s_exc.BadTime, s_agenda.ApptRec, {s_tu.HOUR: -99}, s_tu.MINUTE)
        self.raises(s_exc.BadTime, s_agenda.ApptRec, {}, s_tu.YEAR, -1)
        self.raises(s_exc.BadTime, s_agenda.ApptRec, {'day': 4}, s_tu.YEAR, -1)

        ###############################

        # incunit = None, i.e. one-shot

        self.raises(s_exc.BadTime, s_agenda.ApptRec, {s_tu.DAY: 15})

        # No inc, year req: One shot in 2017.  It is 2018
        now = datetime.datetime(year=2018, month=12, day=3, hour=2, minute=2, tzinfo=tz.utc).timestamp()  # Monday
        ar = s_agenda.ApptRec({s_tu.YEAR: 2017})
        newts = ar.nexttime(now)
        self.eq(newts, 0.0)

        # one shot in 2019.
        ar = s_agenda.ApptRec({s_tu.YEAR: 2019})
        newts = ar.nexttime(now)
        self.eq(newts, datetime.datetime(year=2019, month=12, day=3, hour=2, minute=2, tzinfo=tz.utc).timestamp())

        # It is leap day 2020, one shot next year
        now = datetime.datetime(year=2020, month=2, day=29, hour=2, minute=2, tzinfo=tz.utc).timestamp()  # Monday
        ar = s_agenda.ApptRec({s_tu.YEAR: 2021})
        newts = ar.nexttime(now)
        self.eq(newts, datetime.datetime(year=2021, month=2, day=28, hour=2, minute=2, tzinfo=tz.utc).timestamp())

        # It is leap day 2020, one shot next year, Halloween
        now = datetime.datetime(year=2020, month=2, day=29, hour=2, minute=2, tzinfo=tz.utc).timestamp()  # Monday
        ar = s_agenda.ApptRec({s_tu.DAYOFMONTH: 31, s_tu.YEAR: 2021, s_tu.MONTH: 10})
        newts = ar.nexttime(now)
        self.eq(newts, datetime.datetime(year=2021, month=10, day=31, hour=2, minute=2, tzinfo=tz.utc).timestamp())

        # No inc, month req: One shot in July
        ar = s_agenda.ApptRec({s_tu.MONTH: 7})
        now = datetime.datetime(year=2018, month=12, day=3, hour=2, minute=2, tzinfo=tz.utc).timestamp()  # Monday
        newts = ar.nexttime(now)
        self.eq(newts, datetime.datetime(year=2019, month=7, day=3, hour=2, minute=2, tzinfo=tz.utc).timestamp())

        # No inc, day of month req: one shot on the 3rd of the month at 1:01am
        ar = s_agenda.ApptRec({s_tu.DAYOFMONTH: 3, s_tu.HOUR: 1, s_tu.MINUTE: 1})
        now = datetime.datetime(year=2018, month=12, day=4, hour=2, minute=2, tzinfo=tz.utc).timestamp()  # Monday
        newts = ar.nexttime(now)
        self.eq(newts, datetime.datetime(year=2019, month=1, day=3, hour=1, minute=1, tzinfo=tz.utc).timestamp())

        # No inc, day of week req: One shot next Monday at 1:01 am
        ar = s_agenda.ApptRec({s_tu.DAYOFWEEK: 0, s_tu.HOUR: 1, s_tu.MINUTE: 1})
        now = datetime.datetime(year=2018, month=12, day=31, hour=2, minute=2, tzinfo=tz.utc).timestamp()  # Monday
        newts = ar.nexttime(now)
        self.eq(newts, datetime.datetime(year=2019, month=1, day=7, hour=1, minute=1, tzinfo=tz.utc).timestamp())

        # No inc, hour req:  One shot at 4pm.  It is 5pm, last day of the year.
        ar = s_agenda.ApptRec({s_tu.HOUR: 16})
        now = datetime.datetime(year=2018, month=12, day=31, hour=17, minute=2, tzinfo=tz.utc).timestamp()  # Monday
        newts = ar.nexttime(now)
        self.eq(newts, datetime.datetime(year=2019, month=1, day=1, hour=16, minute=2, tzinfo=tz.utc).timestamp())

        ################

        # incunit = Year

        # Year inc, day of month req:  every year on the 15th of the month
        ar = s_agenda.ApptRec({s_tu.DAYOFMONTH: 15}, s_tu.YEAR, 1)

        # before the target matches the target
        now = datetime.datetime(year=2018, month=3, day=1, tzinfo=tz.utc).timestamp()
        newts = ar.nexttime(now)
        self.eq(newts, datetime.datetime(year=2018, month=3, day=15, tzinfo=tz.utc).timestamp())

        # at the appointment advances
        now = datetime.datetime(year=2018, month=3, day=15, tzinfo=tz.utc).timestamp()
        newts = ar.nexttime(now)
        self.eq(newts, datetime.datetime(year=2019, month=3, day=15, tzinfo=tz.utc).timestamp())

        # past the appointment advances
        now = datetime.datetime(year=2018, month=3, day=15, tzinfo=tz.utc).timestamp()
        newts = ar.nexttime(now)
        self.eq(newts, datetime.datetime(year=2019, month=3, day=15, tzinfo=tz.utc).timestamp())

        # Year inc, month req:  every other year in February
        ar = s_agenda.ApptRec({s_tu.MONTH: 2}, s_tu.YEAR, 2)
        now = datetime.datetime(year=2018, month=12, day=31, hour=17, minute=2, tzinfo=tz.utc).timestamp()  # Monday
        newts = ar.nexttime(now)
        self.eq(newts, datetime.datetime(year=2020, month=2, day=28, hour=17, minute=2, tzinfo=tz.utc).timestamp())

        # Year inc, day of week req:  every other year on a Friday:  not supported
        self.raises(s_exc.BadTime, s_agenda.ApptRec, {s_tu.DAYOFWEEK: 4}, s_tu.YEAR, 2)

        # Year inc, hour req
        ar = s_agenda.ApptRec({s_tu.HOUR: 10}, s_tu.YEAR, 2)
        now = datetime.datetime(year=2018, month=12, day=31, hour=17, minute=2, tzinfo=tz.utc).timestamp()  # Monday
        newts = ar.nexttime(now)
        self.eq(newts, datetime.datetime(year=2020, month=12, day=31, hour=10, minute=2, tzinfo=tz.utc).timestamp())

        #################

        # incunit = Month

        # Month inc, minute req: When the minute hand hits 30 every 14 months
        ar = s_agenda.ApptRec({s_tu.MINUTE: 30}, s_tu.MONTH, 14)
        now = datetime.datetime(year=2017, month=12, day=30, hour=0, minute=30, tzinfo=tz.utc).timestamp()
        newts = ar.nexttime(now)
        self.eq(newts, datetime.datetime(year=2019, month=2, day=28, hour=0, minute=30, tzinfo=tz.utc).timestamp())

        # Month inc, day of month req:  every other month on the 4th
        ar = s_agenda.ApptRec({s_tu.DAYOFMONTH: 4}, s_tu.MONTH, 2)
        now = datetime.datetime(year=2017, month=12, day=30, hour=0, minute=30, tzinfo=tz.utc).timestamp()
        newts = ar.nexttime(now)
        self.eq(newts, datetime.datetime(2018, 2, 4, 0, 30, tzinfo=tz.utc).timestamp())

        # Month inc, day of week fail
        self.raises(s_exc.BadTime, s_agenda.ApptRec, {s_tu.DAYOFWEEK: 4}, s_tu.MONTH, 2)

        ###############

        # incunit = Day

        # Day inc, minute req:  Every day some time when the minute hand hits 30
        ar = s_agenda.ApptRec({s_tu.MINUTE: 30}, s_tu.DAY, 1)
        now = datetime.datetime(year=2019, month=2, day=1, hour=0, minute=2, tzinfo=tz.utc).timestamp()
        newts = ar.nexttime(now)
        self.eq(newts, datetime.datetime(year=2019, month=2, day=1, hour=0, minute=30, tzinfo=tz.utc).timestamp())

        ###############

        # incunit = DayOfWeek

        # Day of week inc, hour req: Every Wednesday at 7pm
        ar = s_agenda.ApptRec({s_tu.HOUR: 7}, s_tu.DAYOFWEEK, 2)
        now = datetime.datetime(year=2018, month=12, day=3, hour=2, minute=2, tzinfo=tz.utc).timestamp()  # Monday
        newts = ar.nexttime(now)
        self.eq(newts, datetime.datetime(year=2018, month=12, day=5, hour=7, minute=2, tzinfo=tz.utc).timestamp())

    async def test_agenda(self):
        MONO_DELT = 1543827303.0
        unixtime = datetime.datetime(year=2018, month=12, day=5, hour=7, minute=0, tzinfo=tz.utc).timestamp()
        sync = asyncio.Event()
        lastquery = None

        def timetime():
            return unixtime

        def looptime():
            return unixtime - MONO_DELT

        async def myeval(query, opts=None):
            nonlocal lastquery
            lastquery = query
            sync.set()
            if 'sleep' in query:
                await asyncio.sleep(60)

            if query == 'badquery':
                raise Exception('test exception')
            return
            yield None

        class FakeAuth(mock.Mock):
            def __init__(self, **kwargs):
                mock.Mock.__init__(self, **kwargs)
                self.users = {}

            def user(self, iden):
                return self.users.get(iden)

            async def addUser(self, username, iden=None):
                if iden is None:
                    iden = s_common.guid()
                user = FakeUser()
                user.name = username
                self.users[iden] = user
                return user

        class FakeUser(mock.Mock):
            def __init__(self, **kwargs):
                mock.Mock.__init__(self, **kwargs)
                self.info = {'locked': False}

            async def setLocked(self, valu):
                self.info['locked'] = valu

        loop = asyncio.get_running_loop()
        with mock.patch.object(loop, 'time', looptime), mock.patch('time.time', timetime), self.getTestDir() as dirn:
            core = mock.Mock()
            core.eval = myeval
            core.slab = await s_lmdbslab.Slab.anit(dirn, map_size=s_t_utils.TEST_MAP_SIZE, readonly=False)
            nexsroot = await s_nexus.NexsRoot.anit(dirn)
            await nexsroot.startup(None)

            def mockgetStormQuery(q):
                # TODO: Remove this and use AsyncMock when Python 3.8+ only
                f = asyncio.Future()
                f.set_result(None)
                return f

            core.getStormQuery = mockgetStormQuery

            db = core.slab.initdb('hive')
            core.hive = await s_hive.SlabHive.anit(core.slab, db=db, nexsroot=nexsroot)

            core.auth = FakeAuth()
            rootiden = 'aaaaa'
            await core.auth.addUser('root', iden=rootiden)

            core.boss = await s_boss.Boss.anit()
            async with await s_agenda.Agenda.anit(core) as agenda:
                agenda.onfini(core.hive)
                agenda.onfini(core.slab)
                agenda.onfini(core.boss)

                await agenda.start()
                await agenda.start()  # make sure it doesn't blow up
                self.eq([], agenda.list())

                # Missing reqs
                cdef = {'creator': rootiden, 'iden': 'fakeiden', 'storm': 'foo'}
                await self.asyncraises(ValueError, agenda.add(cdef))

                # Missing creator
                cdef = {'iden': 'fakeiden', 'storm': 'foo',
                        'reqs': {s_agenda.TimeUnit.MINUTE: 1}}
                await self.asyncraises(ValueError, agenda.add(cdef))

                # Missing storm
                cdef = {'creator': rootiden, 'iden': 'fakeiden',
                        'reqs': {s_agenda.TimeUnit.MINUTE: 1}}
                await self.asyncraises(ValueError, agenda.add(cdef))
                await self.asyncraises(s_exc.NoSuchIden, agenda.get('newp'))

                # Missing incvals
                cdef = {'creator': rootiden, 'iden': 'DOIT', 'storm': '[test:str=doit]',
                        'reqs': {s_agenda.TimeUnit.NOW: True},
                        'incunit': s_agenda.TimeUnit.MONTH}
                await self.asyncraises(ValueError, agenda.add(cdef))

                # Cannot schedule a recurring job with 'now'
                cdef = {'creator': rootiden, 'iden': 'DOIT', 'storm': '[test:str=doit]',
                        'reqs': {s_agenda.TimeUnit.NOW: True},
                        'incunit': s_agenda.TimeUnit.MONTH,
                        'incvals': 1}
                await self.asyncraises(ValueError, agenda.add(cdef))
                await self.asyncraises(s_exc.NoSuchIden, agenda.get('DOIT'))

                # Schedule a one-shot to run immediately
                cdef = {'creator': rootiden, 'iden': 'DOIT', 'storm': '[test:str=doit]',
                        'reqs': {s_agenda.TimeUnit.NOW: True}}
                await agenda.add(cdef)

                # Can't have two with the same iden
                await self.asyncraises(s_exc.DupIden, agenda.add(cdef))

                await sync.wait()  # wait for the query to run
                sync.clear()
                self.eq(lastquery, '[test:str=doit]')
                core.reset_mock()
                lastquery = None

                appts = agenda.list()
                self.len(1, appts)
                self.eq(appts[0][1].startcount, 1)
                self.eq(appts[0][1].nexttime, None)
                await agenda.delete('DOIT')

                # Schedule a one-shot 1 minute from now
                cdef = {'creator': rootiden, 'iden': 'IDEN1', 'storm': '[test:str=foo]',
                        'reqs': {s_agenda.TimeUnit.MINUTE: 1}}
                await agenda.add(cdef)
                await asyncio.sleep(0)  # give the scheduler a shot to wait
                unixtime += 61
                await sync.wait()  # wait for the query to run
                sync.clear()
                self.eq(lastquery, '[test:str=foo]')
                core.reset_mock()
                lastquery = None

                appts = agenda.list()
                self.len(1, appts)
                self.eq(appts[0][1].startcount, 1)
                self.eq(appts[0][1].nexttime, None)

                # Schedule a query to run every Wednesday and Friday at 10:15am
                cdef = {'creator': rootiden, 'iden': 'IDEN2', 'storm': '[test:str=bar]',
                        'reqs': {s_tu.HOUR: 10, s_tu.MINUTE: 15},
                        'incunit': s_agenda.TimeUnit.DAYOFWEEK,
                        'incvals': (2, 4)}
                adef = await agenda.add(cdef)
                guid = adef.get('iden')

                # every 6th of the month at 7am and 8am (the 6th is a Thursday)
                cdef = {'creator': rootiden, 'iden': 'IDEN3', 'storm': '[test:str=baz]',
                        'reqs': {s_tu.HOUR: (7, 8), s_tu.MINUTE: 0, s_tu.DAYOFMONTH: 6},
                        'incunit': s_agenda.TimeUnit.MONTH,
                        'incvals': 1}
                adef = await agenda.add(cdef)
                guid2 = adef.get('iden')

                xmas = {s_tu.DAYOFMONTH: 25, s_tu.MONTH: 12, s_tu.YEAR: 2018}
                lasthanu = {s_tu.DAYOFMONTH: 10, s_tu.MONTH: 12, s_tu.YEAR: 2018}

                # And one-shots for Christmas and last day of Hanukkah of 2018
                cdef = {'creator': rootiden, 'iden': 'IDEN4', 'storm': '#happyholidays',
                        'reqs': (xmas, lasthanu)}
                await agenda.add(cdef)

                await asyncio.sleep(0)
                unixtime += 1
                # Nothing should happen
                self.none(lastquery)

                # Advance to the first event on Wednesday the 5th
                unixtime = datetime.datetime(year=2018, month=12, day=5, hour=10, minute=16, tzinfo=tz.utc).timestamp()
                await sync.wait()
                sync.clear()
                self.eq(lastquery, '[test:str=bar]')

                # Then two on the 6th
                unixtime = datetime.datetime(year=2018, month=12, day=6, hour=7, minute=15, tzinfo=tz.utc).timestamp()
                await sync.wait()
                sync.clear()
                self.eq(lastquery, '[test:str=baz]')
                lastquery = None
                unixtime = datetime.datetime(year=2018, month=12, day=6, hour=8, minute=15, tzinfo=tz.utc).timestamp()
                await sync.wait()
                sync.clear()
                self.eq(lastquery, '[test:str=baz]')

                # Then back to the 10:15 on Friday
                unixtime = datetime.datetime(year=2018, month=12, day=7, hour=10, minute=16, tzinfo=tz.utc).timestamp()
                await sync.wait()
                sync.clear()
                self.eq(lastquery, '[test:str=bar]')

                # Then Dec 10
                unixtime = datetime.datetime(year=2018, month=12, day=10, hour=10, minute=16, tzinfo=tz.utc).timestamp()
                await sync.wait()
                sync.clear()
                self.eq(lastquery, '#happyholidays')

                # Then the Wednesday again
                unixtime = datetime.datetime(year=2018, month=12, day=12, hour=10, minute=16, tzinfo=tz.utc).timestamp()
                await sync.wait()
                sync.clear()
                self.eq(lastquery, '[test:str=bar]')

                # Cancel the Wednesday/Friday appt
                await agenda.delete(guid)
                await self.asyncraises(s_exc.NoSuchIden, agenda.delete(b'1234'))

                # Then Dec 25
                unixtime = datetime.datetime(year=2018, month=12, day=25, hour=10, minute=16, tzinfo=tz.utc).timestamp()
                await sync.wait()
                sync.clear()
                self.eq(lastquery, '#happyholidays')

                # Then Jan 6
                unixtime = datetime.datetime(year=2019, month=1, day=6, hour=10, minute=16, tzinfo=tz.utc).timestamp()
                await sync.wait()
                sync.clear()
                self.eq(lastquery, '[test:str=baz]')

                # Modify the last appointment
                await self.asyncraises(ValueError, agenda.mod(guid2, '', ))
                await agenda.mod(guid2, '#baz')
                self.eq(agenda.appts[guid2].query, '#baz')

                # Delete the other recurring appointment
                await agenda.delete(guid2)

                # Then nothing left scheduled
                self.len(0, agenda.apptheap)

                # Test that isrunning updated, cancelling works
                cdef = {'creator': rootiden, 'iden': 'IDEN5', 'storm': 'inet:ipv4=1 | sleep 120',
                        'reqs': {}, 'incunit': s_agenda.TimeUnit.MINUTE, 'incvals': 1}
                adef = await agenda.add(cdef)
                guid = adef.get('iden')

                unixtime += 60
                await sync.wait()
                sync.clear()
                self.len(1, core.boss.tasks)
                task = next(iter(core.boss.tasks.values()))
                self.eq(task.info.get('query'), 'inet:ipv4=1 | sleep 120')
                self.eq(task.info.get('iden'), guid)

                appt = await agenda.get(guid)
                self.eq(appt.isrunning, True)
                await task.kill()

                appt = await agenda.get(guid)
                self.eq(appt.isrunning, False)
                self.eq(appt.lastresult, 'cancelled')
                await agenda.delete(guid)

                # Test bad queries record exception
                cdef = {'creator': rootiden, 'iden': '#foo', 'storm': 'IDEN',
                        'reqs': {}, 'incunit': s_agenda.TimeUnit.MINUTE,
                        'incvals': 1}
                adef = await agenda.add(cdef)
                guid = adef.get('iden')

                # bypass the API because it would actually syntax check
                agenda.appts[guid].query = 'badquery'
                unixtime += 60
                await sync.wait()
                sync.clear()

                appt = await agenda.get(guid)
                self.eq(appt.isrunning, False)
                self.eq(appt.lastresult, 'raised exception test exception')

                # Test setting the global enable/disable flag
                # reset
                await agenda.delete(guid)
                lastquery = None
                self.len(0, agenda.apptheap)
                agenda._schedtask.cancel()
                agenda._schedtask = agenda.schedCoro(agenda._scheduleLoop())

                visi = await core.auth.addUser('visi', iden='visivisi')

                # schedule a query to run every Wednesday and Friday at 10:15am
                cdef = {'creator': 'visivisi', 'iden': 'IDEN6', 'storm': '[test:str=bar]',
                        'reqs': {s_tu.HOUR: 10, s_tu.MINUTE: 15},
                        'incunit': s_agenda.TimeUnit.DAYOFWEEK,
                        'incvals': (2, 4)}
                adef = await agenda.add(cdef)
                guid = adef.get('iden')

                self.len(1, agenda.apptheap)

                # disable crons and advance time
                agenda.enabled = False

                unixtime = datetime.datetime(year=2019, month=2, day=6, hour=10, minute=16, tzinfo=tz.utc).timestamp()
                nexttime = datetime.datetime(year=2019, month=2, day=8, hour=10, minute=15, tzinfo=tz.utc).timestamp()

                await asyncio.sleep(0)
                self.none(lastquery)

                appt = await agenda.get(guid)
                self.eq(nexttime, appt.nexttime)
                self.true(appt.enabled)
                self.eq(0, appt.startcount)

                # enable crons and advance time
                agenda.enabled = True

                unixtime = datetime.datetime(year=2019, month=2, day=13, hour=10, minute=16, tzinfo=tz.utc).timestamp()
                await sync.wait()
                sync.clear()
                self.eq(lastquery, '[test:str=bar]')
                self.eq(1, appt.startcount)

                cdef = {'creator': rootiden, 'iden': 'IDEN2', 'storm': '[test:str=foo2]',
                        'reqs': {s_agenda.TimeUnit.MINUTE: 1}}
                await agenda.add(cdef)

                # Lock user and advance time
                await visi.setLocked(True)

                with self.getLoggerStream('synapse.lib.agenda', 'locked') as stream:
                    unixtime = datetime.datetime(year=2019, month=2, day=16, hour=10, minute=16,
                                                 tzinfo=tz.utc).timestamp()
                    await sync.wait()
                    sync.clear()
                    self.eq(2, appt.startcount)
                    self.true(stream.wait(0.1))

    async def test_agenda_persistence(self):
        ''' Test we can make/change/delete appointments and they are persisted to storage '''
        with self.getTestDir() as fdir:

            core = mock.Mock()

            def raiseOnBadStorm(q):
                ''' Just enough storm parsing for this test '''
                # TODO: Async this and use AsyncMock when Python 3.8+ only
                f = asyncio.Future()
                if (q[0] == '[') != (q[-1] == ']'):
                    f.set_exception(s_exc.BadSyntax(mesg='mismatched braces'))
                else:
                    f.set_result('all good')
                return f

            core.getStormQuery = raiseOnBadStorm

            async with await s_lmdbslab.Slab.anit(fdir, map_size=100000) as slab:
                db = slab.initdb('hive')
                hive = await s_hive.SlabHive.anit(slab, db=db)
                core.hive = hive

                async with await s_agenda.Agenda.anit(core) as agenda:
                    await agenda.start()
                    # Schedule a query to run every Wednesday and Friday at 10:15am
                    cdef = {'creator': 'visi', 'iden': 'IDEN1', 'storm': '[test:str=bar]',
                            'reqs': {s_tu.HOUR: 10, s_tu.MINUTE: 15},
                            'incunit': s_agenda.TimeUnit.DAYOFWEEK,
                            'incvals': (2, 4)}
                    adef = await agenda.add(cdef)
                    guid1 = adef.get('iden')

                    # every 6th of the month at 7am and 8am (the 6th is a Thursday)
                    cdef = {'creator': 'visi', 'iden': 'IDEN2', 'storm': '[test:str=baz]',
                            'reqs': {s_tu.HOUR: (7, 8), s_tu.MINUTE: 0, s_tu.DAYOFMONTH: 6},
                            'incunit': s_agenda.TimeUnit.MONTH,
                            'incvals': 1}
                    adef = await agenda.add(cdef)

                    guid2 = adef.get('iden')
                    appt = agenda.appts[guid2]
                    appt.lasterrs.append('error happened')
                    await agenda._storeAppt(appt)

                    # Add an appt with an invalid query
                    cdef = {'creator': 'visi', 'iden': 'BAD1', 'storm': '[test:str=',
                            'reqs': {s_tu.HOUR: (7, 8)},
                            'incunit': s_agenda.TimeUnit.MONTH,
                            'incvals': 1}
                    adef = await agenda.add(cdef)
                    badguid1 = adef.get('iden')

                    # Add an appt with a bad version in storage
                    cdef = {'creator': 'visi', 'iden': 'BAD2', 'storm': '[test:str=foo]',
                            'reqs': {s_tu.HOUR: (7, 8)},
                            'incunit': s_agenda.TimeUnit.MONTH,
                            'incvals': 1}
                    adef = await agenda.add(cdef)
                    badguid2 = adef.get('iden')

                    adef['ver'] = 1337
                    full = agenda._hivenode.full + (badguid2,)
                    await agenda.core.hive.set(full, adef)

                    xmas = {s_tu.DAYOFMONTH: 25, s_tu.MONTH: 12, s_tu.YEAR: 2099}
                    lasthanu = {s_tu.DAYOFMONTH: 10, s_tu.MONTH: 12, s_tu.YEAR: 2099}

                    await agenda.delete(guid1)

                    # And one-shots for Christmas and last day of Hanukkah of 2018
                    cdef = {'creator': 'visi', 'iden': 'IDEN3', 'storm': '#happyholidays',
                            'reqs': (xmas, lasthanu)}

                    adef = await agenda.add(cdef)
                    guid3 = adef.get('iden')

                    await agenda.mod(guid3, '#bahhumbug')

                async with await s_agenda.Agenda.anit(core) as agenda:
                    await agenda.start()
                    agenda.enabled = True

                    appts = agenda.list()
                    self.len(3, appts)

                    last_appt = [appt for (iden, appt) in appts if iden == guid3][0]
                    self.eq(last_appt.query, '#bahhumbug')

                    bad_appt = [appt for (iden, appt) in appts if iden == badguid1][0]
                    self.false(bad_appt.enabled)

                    self.len(0, [appt for (iden, appt) in appts if iden == badguid2])

    async def test_cron_perms(self):

        async with self.getTestCore() as core:

            visi = await core.auth.addUser('visi')
            newb = await core.auth.addUser('newb')
            async with core.getLocalProxy(user='visi') as proxy:

                cdef = {'storm': 'inet:ipv4', 'reqs': {'hour': 2}}
                with self.raises(s_exc.AuthDeny):
                    await proxy.addCronJob(cdef)

                await visi.addRule((True, ('cron', 'add')))
                cron0 = await proxy.addCronJob(cdef)
                cron0_iden = cron0.get('iden')

                cdef = {'storm': 'inet:ipv6', 'reqs': {'hour': 2}}
                cron1 = await proxy.addCronJob(cdef)
                cron1_iden = cron1.get('iden')

                await proxy.delCronJob(cron0_iden)

                cdef = {'storm': '[test:str=foo]', 'reqs': {'now': True},
                        'incunit': 'month',
                        'incvals': 1}
                await self.asyncraises(s_exc.BadConfValu, proxy.addCronJob(cdef))

            async with core.getLocalProxy(user='newb') as proxy:

                with self.raises(s_exc.AuthDeny):
                    await proxy.delCronJob(cron1_iden)

                self.eq(await proxy.listCronJobs(), ())
                await newb.addRule((True, ('cron', 'get')))
                self.len(1, await proxy.listCronJobs())

                with self.raises(s_exc.AuthDeny):
                    await proxy.disableCronJob(cron1_iden)

                await newb.addRule((True, ('cron', 'set')))
                self.none(await proxy.disableCronJob(cron1_iden))

                await newb.addRule((True, ('cron', 'del')))
                await proxy.delCronJob(cron1_iden)

    async def test_agenda_stop(self):

        async with self.getTestCore() as core:
            await core.callStorm('$lib.queue.add(foo)')
            await core.callStorm('cron.add --minute +1 { $lib.queue.get(foo).put((99)) $lib.time.sleep(10) }')
            appts = core.agenda.list()
            await core.agenda._execute(appts[0][1])
            self.eq((0, 99), await core.callStorm('return($lib.queue.get(foo).get())'))

            appts[0][1].creator = 'fakeuser'
            await core.agenda._execute(appts[0][1])
            self.eq(appts[0][1].lastresult, 'Failed due to unknown user')

            await core.agenda.stop()
            self.false(core.agenda.enabled)

    async def test_agenda_custom_view(self):

        async with self.getTestCoreAndProxy() as (core, prox):
            # no existing view
            await core.callStorm('$lib.queue.add(testq)')
            defview = core.getView()
            fakeiden = hashlib.md5(defview.iden.encode('utf-8')).hexdigest()
            opts = {'vars': {'fakeiden': fakeiden}}
            with self.raises(s_exc.NoSuchView):
                await prox.callStorm('cron.add --view $fakeiden --minute +2 { $lib.queue.get(testq).put((43)) }', opts=opts)

            fail = await core.auth.addUser('fail')

            # can't move a thing that doesn't exist
            with self.raises(s_exc.StormRuntimeError):
                await core.callStorm('cron.move $fakeiden $fakeiden', opts=opts)

            with self.raises(s_exc.NoSuchIden):
                await core.moveCronJob(fail.iden, 'NoSuchCronJob', defview.iden)

            with self.raises(s_exc.NoSuchIden):
                await core.agenda.move('StillDoesNotExist', defview.iden)

            # make a new view
            ldef = await core.addLayer()
            newlayr = core.getLayer(ldef.get('iden'))
            newview = await core.callStorm(f'return($lib.view.add(({newlayr.iden},)).iden)')

            # no perms to write to that view
            asfail = {'user': fail.iden, 'vars': {'newview': newview}}
            with self.raises(s_exc.AuthDeny):
                await prox.callStorm('cron.add --view $newview --minute +2 { $lib.queue.get(testq).put(lolnope) }', opts=asfail)

            # and just to be sure
            msgs = await core.stormlist('cron.list')
            self.stormIsInPrint('No cron jobs found', msgs)

            # no --view means it goes in the default view for the user, which fail doesn't have rights to
            with self.raises(s_exc.AuthDeny):
                await core.callStorm('cron.add --minute +1 { $lib.queue.get(testq).put((44)) }', opts=asfail)

            # let's give fail permissions to do some things, but not in our super special view (fail is missing
            # the view read perm for the special view)
            await fail.addRule((True, ('cron', 'add')))

            # But we should still fail on this:
            with self.raises(s_exc.AuthDeny):
                await core.callStorm('cron.add --view $newview --minute +2 { $lib.queue.get(testq).put(lolnope)}', opts=asfail)

            # and again, just to be sure
            msgs = await core.stormlist('cron.list')
            self.stormIsInPrint('No cron jobs found', msgs)

            # Now let's give him perms to do things
            await fail.addRule((True, ('view', 'read')), gateiden=newview)
            await fail.addRule((True, ('queue', 'get')), gateiden='queue:testq')
            await fail.addRule((True, ('queue', 'put')), gateiden='queue:testq')
            await fail.addRule((True, ('node', 'add')))
            await fail.addRule((True, ('cron', 'get')))

            # but should work on the default view
            opts = {'user': fail.iden, 'view': defview.iden, 'vars': {'defview': defview.iden}}
            await prox.callStorm('cron.at --view $defview --minute +1 { $lib.queue.get(testq).put((44)) }', opts=opts)

            jobs = await core.callStorm('return($lib.cron.list())')
            self.len(1, jobs)
            self.eq(defview.iden, jobs[0]['view'])

            core.agenda._addTickOff(60)
            retn = await core.callStorm('return($lib.queue.get(testq).get())', opts=asfail)
            self.eq((0, 44), retn)

            await core.callStorm('cron.del $croniden', opts={'vars': {'croniden': jobs[0]['iden']}})
            await core.callStorm('$lib.queue.get(testq).cull(0)')

            opts = {'vars': {'newview': newview}}
            await prox.callStorm('cron.add --minute +1 --view $newview { [test:guid=$lib.guid()] | $lib.queue.get(testq).put($node) }', opts=opts)

            jobs = await core.callStorm('return($lib.cron.list())')
            self.len(1, jobs)
            self.eq(newview, jobs[0]['view'])

            core.agenda._addTickOff(60)
            retn = await core.callStorm('return($lib.queue.get(testq).get())')
            await core.callStorm('$lib.queue.get(testq).cull(1)')

            # That node had better have been made in the new view
            guidnode = await core.nodes('test:guid', opts={'view': newview})
            self.len(1, guidnode)
            self.eq(('test:guid', retn[1]), guidnode[0].ndef)

            # and definitely not in the base view
            nonode = await core.nodes('test:guid', opts={'view': defview.iden})
            self.len(0, nonode)

            # no permission yet
            opts = {'user': fail.iden, 'vars': {'croniden': jobs[0]['iden'], 'viewiden': defview.iden}}
            with self.raises(s_exc.StormRuntimeError):
                await core.callStorm('cron.move $croniden $viewiden', opts=opts)

            await fail.addRule((True, ('cron', 'set')))
            # try and fail to move to a view that doesn't exist
            opts = {'user': fail.iden, 'vars': {'croniden': jobs[0]['iden'], 'viewiden': fakeiden}}
            with self.raises(s_exc.NoSuchView):
                await core.callStorm('cron.move $croniden $viewiden', opts=opts)

            croniden = jobs[0]['iden']
            # now to test that we can move from the new layer to the base layer
            opts = {'user': fail.iden, 'vars': {'croniden': croniden, 'viewiden': defview.iden}}
            await core.callStorm('cron.move $croniden $viewiden', opts=opts)

            jobs = await core.callStorm('return($lib.cron.list())')
            self.len(1, jobs)
            self.eq(defview.iden, jobs[0]['view'])

            # moving to the same view shouldn't do much
            await core.moveCronJob(fail.iden, croniden, defview.iden)

            samejobs = await core.callStorm('return($lib.cron.list())')
            self.len(1, jobs)
            self.eq(jobs, samejobs)

            core.agenda._addTickOff(60)
            retn = await core.callStorm('return($lib.queue.get(testq).get())', opts=asfail)
            await core.callStorm('$lib.queue.get(testq).cull(2)')

            node = await core.nodes('test:guid', opts={'view': defview.iden})
            self.len(1, node)
            self.eq(('test:guid', retn[1]), node[0].ndef)
            self.ne(guidnode[0].ndef, node[0].ndef)

            # reach in, monkey with the view a bit
            appt = core.agenda.appts.get(croniden)
            appt.view = "ThisViewStillDoesntExist"
            await core.agenda._execute(appt)
            self.eq(appt.lastresult, 'Failed due to unknown view')

            await core.callStorm('cron.del $croniden', opts={'vars': {'croniden': croniden}})

            opts = {'vars': {'newview': newview}}
            await prox.callStorm('cron.at --now --view $newview { [test:str=gotcha] | $lib.queue.get(testq).put($node) }', opts=opts)
            retn = await core.callStorm('return($lib.queue.get(testq).get())', opts=asfail)
            self.eq((3, 'gotcha'), retn)
            await core.callStorm('$lib.queue.get(testq).cull(3)')
            atjob = await core.callStorm('return($lib.cron.list())')
            self.len(1, atjob)
            self.eq(atjob[0]['view'], newview)

            nodes = await core.nodes('test:str', opts={'view': defview.iden})
            self.len(0, nodes)
            nodes = await core.nodes('test:str', opts={'view': newview})
            self.len(1, nodes)

            croniden = atjob[0]['iden']
            await core.callStorm('cron.del $croniden', opts={'vars': {'croniden': croniden}})

            await prox.callStorm('cron.at --now { [test:int=97] | $lib.queue.get(testq).put($node) }')
            retn = await core.callStorm('return($lib.queue.get(testq).get())', opts=asfail)
            await core.callStorm('$lib.queue.get(testq).cull(4)')
            self.eq((4, 97), retn)

            nodes = await core.nodes('test:int=97', opts={'view': defview.iden})
            self.len(1, nodes)
            nodes = await core.nodes('test:int=97', opts={'view': newview})
            self.len(0, nodes)



import asyncio
import collections

import synapse.exc as s_exc
import synapse.common as s_common
import synapse.lib.base as s_base

class AQueue(s_base.Base):
    '''
    An async queue with chunk optimized sync compatible consumer.
    '''
    async def __anit__(self):

        await s_base.Base.__anit__(self)

        self.fifo = []
        self.event = asyncio.Event()
        self.onfini(self.event.set)

    def put(self, item):
        '''
        Add an item to the queue.
        '''
        if self.isfini:
            return False

        self.fifo.append(item)

        if len(self.fifo) == 1:
            self.event.set()

        return True

    async def slice(self):

        # sync interface to the async queue
        if len(self.fifo) == 0:
            await self.event.wait()

        retn = list(self.fifo)
        self.fifo.clear()
        self.event.clear()
        return retn

class Queue:
    '''
    An asyncio Queue with batch methods and graceful close.
    '''
    def __init__(self, maxsize=None):
        self.q = asyncio.Queue(maxsize=maxsize)
        self.closed = False

    async def close(self):
        await self.q.put(s_common.novalu)
        self.closed = True

    async def put(self, item):

        if self.closed:
            mesg = 'The Queue has been closed.'
            raise s_exc.BadArg(mesg=mesg)

        await self.q.put(item)

    async def size(self):
        size = self.q.qsize()
        if self.closed:
            size -= 1
        return size

    async def puts(self, items):

        if self.closed:
            mesg = 'The Queue has been closed.'
            raise s_exc.BadArg(mesg=mesg)

        for item in items:
            await self.q.put(item)

    async def slice(self, size=1000):

        if self.closed and self.q.qsize() == 0:
            return None

        items = []

        item = await self.q.get()
        if item is s_common.novalu:
            return None

        items.append(item)

        size -= 1

        for i in range(min(size, self.q.qsize())):

            item = await self.q.get()
            if item is s_common.novalu:
                break

            items.append(item)

        return items

    async def slices(self, size=1000):

        while True:
            items = await self.slice(size=size)
            if items is None:
                return

            yield items

class Window(s_base.Base):
    '''
    A Queue like object which yields added items.  If the queue ever reaches
    its maxsize, it will be fini()d.  On fini(), the Window will continue to
    yield results until empty and then return.
    '''
    async def __anit__(self, maxsize=None):
        await s_base.Base.__anit__(self)
        self.maxsize = maxsize
        self.event = asyncio.Event()

        self.linklist = collections.deque()

        async def fini():
            self.event.set()

        self.onfini(fini)

    async def __aiter__(self):

        while True:

            if self.linklist:
                yield self.linklist.popleft()
                continue

            if self.isfini:
                return

            self.event.clear()
            await self.event.wait()

    async def put(self, item):
        '''
        Add a single item to the Window.
        '''
        if self.isfini:
            return False

        self.linklist.append(item)
        self.event.set()

        if self.maxsize is not None and len(self.linklist) >= self.maxsize:
            await self.fini()

        return True

    async def puts(self, items):
        '''
        Add multiple items to the window.
        '''
        if self.isfini:
            return False

        self.linklist.extend(items)
        self.event.set()

        if self.maxsize is not None and len(self.linklist) >= self.maxsize:
            await self.fini()

        return True



import collections

import synapse.exc as s_exc
import synapse.common as s_common

import synapse.lib.node as s_node

import synapse.tests.utils as s_t_utils
from synapse.tests.utils import alist

class NodeTest(s_t_utils.SynTest):

    async def test_pack(self):
        form = 'test:str'
        valu = 'cool'
        props = {'tick': 12345}

        async with self.getTestCore() as core:

            await core.addTagProp('score', ('int', {}), {})
            await core.addTagProp('note', ('str', {'lower': True, 'strip': 'True'}), {})

            async with await core.snap() as snap:

                node = await snap.addNode(form, valu, props=props)
                await node.setTagProp('foo', 'score', 10)
                await node.setTagProp('foo', 'note', " This is a really cool tag! ")

                iden, info = node.pack()
                self.eq(iden, ('test:str', 'cool'))
                self.eq(info.get('tags'), {'foo': (None, None)})
                self.eq(info.get('tagprops'), {'foo': {'score': 10, 'note': 'this is a really cool tag!'}})
                props = {k: v for (k, v) in info.get('props', {}).items() if not k.startswith('.')}
                self.eq(props, {'tick': 12345})

                iden, info = node.pack(dorepr=True)
                self.eq(iden, ('test:str', 'cool'))
                self.eq(info.get('tags'), {'foo': (None, None)})
                props = {k: v for (k, v) in info.get('props', {}).items() if not k.startswith('.')}
                self.eq(props, {'tick': 12345})
                self.eq(info.get('repr'), None)
                reprs = {k: v for (k, v) in info.get('reprs', {}).items() if not k.startswith('.')}
                self.eq(reprs, {'tick': '1970/01/01 00:00:12.345'})
                tagpropreprs = info.get('tagpropreprs')
                self.eq(tagpropreprs, {'foo': {'score': '10'}})

                # Set a property on the node which is extra model and pack it.
                # This situation can be encountered in a multi-layer situation
                # where one Cortex can have model knowledge and set props
                # that another Cortex (sitting on top of the first one) lifts
                # a node which has props the second cortex doens't know about.
                node.props['.newp'] = 1
                node.props['newp'] = (2, 3)
                node.tagprops['foo']['valu'] = 10
                iden, info = node.pack(dorepr=True)
                props, reprs = info.get('props'), info.get('reprs')
                tagprops, tagpropreprs = info.get('tagprops'), info.get('tagpropreprs')
                self.eq(props.get('.newp'), 1)
                self.eq(props.get('newp'), (2, 3))
                self.eq(tagprops, {'foo': {'score': 10, 'note': 'this is a really cool tag!', 'valu': 10}})

                # without model knowledge it is impossible to repr a value so it should
                # *not* be in the repr dict
                self.none(reprs.get('newp'))
                self.none(reprs.get('.newp'))
                self.eq(tagpropreprs, {'foo': {'score': '10'}})

    async def test_set(self):
        form = 'test:str'
        valu = 'cool'
        props = {'tick': 12345}

        async with self.getTestCore() as core:
            async with await core.snap() as snap:
                self.true(snap.strict)  # Following assertions based on snap.strict being true
                node = await snap.addNode(form, valu, props=props)

                self.false(await node.set('tick', 12345))
                self.true(await node.set('tick', 123456))
                await self.asyncraises(s_exc.NoSuchProp, node.set('notreal', 12345))

                ronode = await snap.addNode('test:comp', (1, 's'))
                await self.asyncraises(s_exc.ReadOnlyProp, ronode.set('hehe', 2))
                snap.strict = False
                self.false(await ronode.set('hehe', 3))
                snap.strict = True

    async def test_has(self):
        form = 'test:str'
        valu = 'cool'
        props = {'tick': 12345}

        async with self.getTestCore() as core:
            async with await core.snap() as snap:
                node = await snap.addNode(form, valu, props=props)

                self.true(node.has('tick'))
                self.true(node.has('.created'))
                self.false(node.has('nope'))
                self.false(node.has('.nope'))

    async def test_get(self):
        form = 'test:str'
        valu = 'cool'
        props = {'tick': 12345}

        async with self.getTestCore() as core:
            async with await core.snap() as snap:
                node = await snap.addNode(form, valu, props=props)
                await node.addTag('cool', valu=(1, 2))

                self.eq(node.get('tick'), 12345)
                self.none(node.get('nope'))

                self.eq(node.get('#cool'), (1, 2))
                self.none(node.get('#newp'))

    async def test_pop(self):
        form = 'test:str'
        valu = 'cool'
        props = {'tick': 12345}

        async with self.getTestCore() as core:
            async with await core.snap() as snap:
                node = await snap.addNode(form, valu, props=props)
                await node.addTag('cool', valu=(1, 2))

                await self.asyncraises(s_exc.NoSuchProp, node.pop('nope'))
                snap.strict = False
                self.false(await node.pop('nope'))
                snap.strict = True

                ronode = await snap.addNode('test:comp', (1, 's'))
                await self.asyncraises(s_exc.ReadOnlyProp, ronode.pop('hehe'))
                snap.strict = False
                self.false(await ronode.pop('hehe'))
                snap.strict = True

    async def test_repr(self):
        async with self.getTestCore() as core:
            async with await core.snap() as snap:

                form = 'test:str'
                valu = 'cool'
                props = {'tick': 12345}
                node = await snap.addNode(form, valu, props=props)
                self.eq('cool', node.repr())
                self.eq(node.repr('tick'), '1970/01/01 00:00:12.345')

                form = 'test:threetype'
                valu = 'cool'
                node = await snap.addNode(form, valu)
                self.eq(node.repr(), '3')
                reprs = {k: v for (k, v) in node.reprs().items() if not k.startswith('.')}
                self.eq(reprs.get('three'), '3')

    async def test_tags(self):
        form = 'test:str'
        valu = 'cool'
        props = {'tick': 12345}

        async with self.getTestCore() as core:
            async with await core.snap() as snap:
                self.true(snap.strict)

                node = await snap.addNode(form, valu, props=props)

                # Add a tag
                await node.addTag('cool', valu=(1, 2))
                self.eq(node.getTag('cool'), (1, 2))
                await node.addTag('cool', valu=(1, 2))  # Add again
                self.eq(node.getTag('cool'), (1, 2))
                await node.addTag('cool', valu=(1, 3))  # Add again with different valu
                self.eq(node.getTag('cool'), (1, 3))
                await node.addTag('cool', valu=(-5, 0))  # Add again with different valu
                self.eq(node.getTag('cool'), (-5, 3))  # merges...

                self.true(node.hasTag('cool'))
                self.true(node.hasTag('#cool'))
                self.false(node.hasTag('notcool'))
                self.false(node.hasTag('#notcool'))

                # Demonstrate that valu is only applied at the level that addTag is called
                await node.addTag('cool.beans.abc', valu=(1, 8))
                self.eq(node.getTag('cool.beans.abc'), (1, 8))
                self.eq(node.getTag('cool.beans'), (None, None))

                await self.asyncraises(s_exc.NoSuchProp, node.pop('nope'))
                snap.strict = False
                self.false(await node.pop('nope'))
                snap.strict = True

    async def test_helpers(self):
        form = 'test:str'
        valu = 'cool'
        props = {'tick': 12345,
                 'hehe': 'hehe',
                 }
        tval = (None, None)

        async with self.getTestCore() as core:
            await core.addTagProp('score', ('int', {}), {})
            await core.addTagProp('note', ('str', {'lower': True, 'strip': 'True'}), {})
            async with await core.snap() as snap:
                node = await snap.addNode(form, valu, props=props)
                await node.addTag('test.foo.bar.duck', tval)
                await node.addTag('test.foo.baz', tval)
                await node.addTag('test.foo.time', ('2016', '2019'))
                await node.addTag('test.foo', ('2015', '2017'))
                await node.setTagProp('test', 'score', 0)
                await node.setTagProp('test', 'note', 'Words')

                pode = node.pack(dorepr=True)

                node2 = await snap.addNode('test:int', '1234')
                pode2 = node2.pack(dorepr=True)

        self.eq(s_node.ndef(pode), ('test:str', 'cool'))
        self.eq(s_node.reprNdef(pode), ('test:str', 'cool'))
        self.eq(s_node.ndef(pode2), ('test:int', 1234))
        self.eq(s_node.reprNdef(pode2), ('test:int', '1234'))

        e = 'bf1198c5f28dae61d595434b0788dd6f7206b1e62d06b0798e012685f1abc85d'
        self.eq(s_node.iden(pode), e)

        self.true(s_node.tagged(pode, 'test'))
        self.true(s_node.tagged(pode, '#test.foo.bar'))
        self.true(s_node.tagged(pode, 'test.foo.bar.duck'))
        self.false(s_node.tagged(pode, 'test.foo.bar.newp'))

        self.len(3, s_node.tags(pode, leaf=True))
        self.len(5, s_node.tagsnice(pode))
        self.len(6, s_node.tags(pode))
        self.eq(s_node.reprTag(pode, '#test.foo.bar'), '')
        self.eq(s_node.reprTag(pode, '#test.foo.time'), '(2016/01/01 00:00:00.000, 2019/01/01 00:00:00.000)')
        self.none(s_node.reprTag(pode, 'test.foo.newp'))

        self.eq(s_node.prop(pode, 'hehe'), 'hehe')
        self.eq(s_node.prop(pode, 'tick'), 12345)
        self.eq(s_node.prop(pode, ':tick'), 12345)
        self.eq(s_node.prop(pode, 'test:str:tick'), 12345)
        self.none(s_node.prop(pode, 'newp'))

        self.eq(s_node.reprProp(pode, 'hehe'), 'hehe')
        self.eq(s_node.reprProp(pode, 'tick'), '1970/01/01 00:00:12.345')
        self.eq(s_node.reprProp(pode, ':tick'), '1970/01/01 00:00:12.345')
        self.eq(s_node.reprProp(pode, 'test:str:tick'), '1970/01/01 00:00:12.345')
        self.none(s_node.reprProp(pode, 'newp'))

        self.eq(s_node.reprTagProps(pode, 'test'),
                [('note', 'words'), ('score', '0')])
        self.eq(s_node.reprTagProps(pode, 'newp'), [])
        self.eq(s_node.reprTagProps(pode, 'test.foo'), [])

        props = s_node.props(pode)
        self.isin('.created', props)
        self.isin('tick', props)
        self.notin('newp', props)

    async def test_storm(self):

        async with self.getTestCore() as core:
            async with await core.snap() as snap:
                query = await snap.core.getStormQuery('')
                async with snap.getStormRuntime(query) as runt:
                    node = await snap.addNode('test:comp', (42, 'lol'))
                    nodepaths = await alist(node.storm(runt, '-> test:int'))
                    self.len(1, nodepaths)
                    self.eq(nodepaths[0][0].ndef, ('test:int', 42))

                    nodepaths = await alist(node.storm(runt, '-> test:int [:loc=$foo]', opts={'vars': {'foo': 'us'}}))
                    self.eq(nodepaths[0][0].props.get('loc'), 'us')

                    path = nodepaths[0][1].fork(node)  # type: s_node.Path
                    path.vars['zed'] = 'ca'

                    # Path present, opts not present
                    nodes = await alist(node.storm(runt, '-> test:int [:loc=$zed] $bar=$foo', path=path))
                    self.eq(nodes[0][0].props.get('loc'), 'ca')
                    # path is not updated due to frame scope
                    self.none(path.vars.get('bar'), 'us')

                    # Path present, opts present but no opts['vars']
                    nodes = await alist(node.storm(runt, '-> test:int [:loc=$zed] $bar=$foo', opts={}, path=path))
                    self.eq(nodes[0][0].props.get('loc'), 'ca')
                    # path is not updated due to frame scope
                    self.none(path.vars.get('bar'))

                    # Path present, opts present with vars
                    nodes = await alist(node.storm(runt, '-> test:int [:loc=$zed] $bar=$baz',
                                                   opts={'vars': {'baz': 'ru'}},
                                                   path=path))
                    self.eq(nodes[0][0].props.get('loc'), 'ca')
                    # path is not updated due to frame scope
                    self.none(path.vars.get('bar'))

                    # Path can push / pop vars in frames
                    self.eq(path.getVar('key'), s_common.novalu)
                    self.len(0, path.frames)
                    path.initframe({'key': 'valu'})
                    self.len(1, path.frames)
                    self.eq(path.getVar('key'), 'valu')
                    path.finiframe()
                    self.len(0, path.frames)
                    self.eq(path.getVar('key'), s_common.novalu)

                    # Path can push / pop a runt as well
                    # This example is *just* a test example to show the variable movement,
                    # not as actual runtime movement..
                    path.initframe({'key': 'valu'})
                    self.eq(path.getVar('key'), 'valu')
                    path.finiframe()
                    self.eq(path.getVar('key'), s_common.novalu)

                    # Path clone() creates a fully independent Path object
                    pcln = path.clone()
                    # Ensure that path vars are independent
                    await pcln.setVar('bar', 'us')
                    self.eq(pcln.getVar('bar'), 'us')
                    self.eq(path.getVar('bar'), s_common.novalu)
                    # Ensure the path nodes are independent
                    self.eq(len(pcln.nodes), len(path.nodes))
                    pcln.nodes.pop(-1)
                    self.ne(len(pcln.nodes), len(path.nodes))

                    # push a frame and clone it - ensure clone mods do not
                    # modify the original path
                    path.initframe({'key': 'valu'})
                    self.len(1, path.frames)
                    pcln = path.clone()
                    self.len(1, pcln.frames)
                    self.eq(path.getVar('key'), 'valu')
                    self.eq(pcln.getVar('key'), 'valu')
                    pcln.finiframe()
                    path.finiframe()
                    await pcln.setVar('bar', 'us')
                    self.eq(pcln.getVar('bar'), 'us')
                    self.eq(path.getVar('bar'), s_common.novalu)
                    self.eq(pcln.getVar('key'), s_common.novalu)
                    self.eq(path.getVar('key'), s_common.novalu)

                    # Check that finiframe without frames resets vars
                    path.finiframe()
                    self.len(0, path.frames)
                    self.eq(s_common.novalu, path.getVar('bar'))

        # Ensure that path clone() behavior in storm is as expected
        # with a real-world style test..
        async with self.getTestCore() as core:
            await core.nodes('[test:int=1 test:int=2]')
            q = '''test:int
            $x = $node.value()
            for $var in (1, 2) { } // The forloop here is used as a node multiplier
            $x = $( $x + 1 )
            $lib.fire(test, valu=$node.value(), x=$x)
            -test:int'''
            msgs = await core.stormlist(q)
            data = collections.defaultdict(set)
            for m in msgs:
                if m[0] == 'storm:fire':
                    for k, v in m[1].get('data').items():
                        data[k].add(v)
            self.eq(dict(data),
                    {'valu': {1, 2}, 'x': {2, 3}})

    async def test_node_repr(self):

        async with self.getTestCore() as core:

            nodes = await core.nodes('[ inet:ipv4=1.2.3.4 :loc=us ]')
            self.len(1, nodes)

            node = nodes[0]

            self.eq('1.2.3.4', nodes[0].repr())

            self.eq('us', node.repr('loc'))

            with self.raises(s_exc.NoSuchProp):
                node.repr('newp')

            self.none(node.repr('dns:rev'))

    async def test_node_data(self):
        async with self.getTestCore() as core:
            nodes = await core.nodes('[ inet:ipv4=1.2.3.4 :loc=us ]')
            self.len(1, nodes)

            node = nodes[0]

            self.none(await node.getData('foo'))

            await node.setData('foo', 123)
            self.eq(123, await node.getData('foo'))

            await node.setData('foo', 123)
            self.eq(123, await node.getData('foo'))

            await node.setData('bar', (4, 5, 6))
            self.eq((4, 5, 6), await node.getData('bar'))
            self.eq(123, await node.getData('foo'))

            self.eq([('foo', 123), ('bar', (4, 5, 6))], await alist(node.iterData()))

            self.eq(123, await node.popData('foo'))
            self.none(await node.getData('foo'))
            self.none(await node.popData('foo'))

            self.eq((4, 5, 6), await node.getData('bar'))

            await node.delete()
            nodes = await core.nodes('[ inet:ipv4=1.2.3.4 :loc=us ]')
            node = nodes[0]

            self.none(await node.getData('foo'))
            self.none(await node.getData('bar'))

    async def test_node_tagprops(self):
        async with self.getTestCore() as core:
            await core.addTagProp('score', ('int', {}), {})
            await core.addTagProp('limit', ('int', {}), {})
            nodes = await core.nodes('[ test:int=10 ]')
            node = nodes[0]

            self.eq(node.tagprops, {})
            await node.setTagProp('foo.test', 'score', 20)
            await node.setTagProp('foo.test', 'limit', 1000)
            self.eq(node.tagprops, {'foo.test': {'score': 20, 'limit': 1000}})

            await node.delTagProp('foo.test', 'score')
            self.eq(node.tagprops, {'foo.test': {'limit': 1000}})

            await node.setTagProp('foo.test', 'score', 50)
            node.tagprops['foo.test'].pop('score')
            await node.delTagProp('foo.test', 'score')
            self.eq(node.tagprops, {'foo.test': {'limit': 1000}})
            node.tagprops['foo.test'].pop('limit')
            self.eq(node.tagprops, {'foo.test': {}})

    async def test_node_edges(self):

        async with self.getTestCore() as core:
            nodes = await core.nodes('[inet:ipv4=1.2.3.4 inet:ipv4=5.5.5.5]')
            with self.raises(s_exc.BadArg):
                await nodes[0].addEdge('foo', 'bar')
            with self.raises(s_exc.BadArg):
                await nodes[0].delEdge('foo', 'bar')



import os
import asyncio

import synapse.exc as s_exc

import synapse.lib.coro as s_coro
import synapse.lib.lmdbslab as s_lmdbslab
import synapse.lib.slabseqn as s_slabseqn

import synapse.tests.utils as s_t_utils

class SlabSeqn(s_t_utils.SynTest):

    def chk_size(self, seqn):
        self.eq(seqn.stat()['entries'], seqn.size)

    async def test_slab_seqn(self):

        with self.getTestDir() as dirn:

            path = os.path.join(dirn, 'test.lmdb')
            slab = await s_lmdbslab.Slab.anit(path, map_size=1000000)

            seqn = s_slabseqn.SlabSeqn(slab, 'seqn:test')
            self.chk_size(seqn)

            self.eq(seqn.nextindx(), 0)
            items = ('foo', 10, 20)
            seqn.save(items)
            retn = tuple(seqn.iter(0))
            self.eq(retn, ((0, 'foo'), (1, 10), (2, 20)))
            self.chk_size(seqn)

            self.raises(s_exc.NotMsgpackSafe, seqn.save, ({'set'},))
            retn = tuple(seqn.iter(0))
            self.eq(retn, ((0, 'foo'), (1, 10), (2, 20)))

            self.eq(seqn.nextindx(), 3)

            await slab.fini()

            # Reopen the seqn and continue where we left off
            slab = await s_lmdbslab.Slab.anit(path, map_size=1000000)

            seqn = s_slabseqn.SlabSeqn(slab, 'seqn:test')
            self.eq(seqn.index(), 3)
            self.chk_size(seqn)

            self.eq(seqn.nextindx(), 3)
            seqn.save(items)

            retn = tuple(seqn.iter(0))
            self.eq(retn, ((0, 'foo'), (1, 10), (2, 20),
                           (3, 'foo'), (4, 10), (5, 20)))
            self.eq(seqn.nextindx(), 6)

            # We can also start in the middle of the sequence
            retn = tuple(seqn.iter(4))
            self.eq(retn, ((4, 10), (5, 20)))

            # iterating past the end yields nothing
            retn = tuple(seqn.iter(100))
            self.eq(retn, ())

            evnt = seqn.getOffsetEvent(4)
            self.true(evnt.is_set())

            evnt1 = seqn.getOffsetEvent(8)
            evnt2 = seqn.getOffsetEvent(9)
            evnt3 = seqn.getOffsetEvent(8)

            seqn.save(items)
            retn = tuple(seqn.iter(0))
            self.len(9, retn)
            self.chk_size(seqn)

            self.eq('foo', seqn.getByIndxByts(b'\x00' * 8))

            self.true(evnt1.is_set())
            self.true(await seqn.waitForOffset(8, timeout=0.5))
            self.false(evnt2.is_set())
            self.false(await seqn.waitForOffset(9, timeout=0.1))
            self.true(evnt3.is_set())

            state = None

            started = asyncio.Event()

            async def taskloop():
                nonlocal state
                state = 'started'
                started.set()
                state = await seqn.waitForOffset(9, timeout=5)

            task = asyncio.get_running_loop().create_task(taskloop())
            self.true(await s_coro.event_wait(started, 2))
            self.eq(state, 'started')

            seqn.add('bar')
            self.true(evnt2.is_set())

            self.true(state)
            await task

            self.eq((0, 'foo'), seqn.pop(0))
            self.none(seqn.pop(0))
            self.chk_size(seqn)

            async def getter():
                retn = []
                async for item in seqn.gets(8):
                    if item[1] is None:
                        return retn
                    retn.append(item)
                return retn

            task = slab.schedCoro(getter())
            await asyncio.sleep(0)
            seqn.add(None)

            self.eq(((8, 20), (9, 'bar')), await asyncio.wait_for(task, timeout=3))

            await seqn.cull(8)
            self.chk_size(seqn)

            self.eq(((9, 'bar'), (10, None)), [x async for x in seqn.gets(8, wait=False)])

            # overwrite existing
            seqn.add('baz', indx=9)
            self.chk_size(seqn)

            # no oldv for smaller indx
            seqn.add('bam', indx=8)
            self.chk_size(seqn)

            # append indx
            seqn.add('faz', indx=15)
            self.chk_size(seqn)

            await seqn.cull(14)
            self.chk_size(seqn)

            seqn.trim(0)
            self.chk_size(seqn)

            await slab.fini()



import json

import synapse.exc as s_exc

import synapse.tests.utils as s_test

logname = 'synapse.storm.log'


class LogTest(s_test.SynTest):

    async def test_stormlib_log(self):

        async with self.getTestCore() as core:
            # Raw message
            with self.getAsyncLoggerStream(logname, 'debug message') as stream:
                await core.callStorm('$lib.log.debug("debug message")')
                self.true(await stream.wait(6))
            with self.getAsyncLoggerStream(logname, 'info message') as stream:
                await core.callStorm('$lib.log.info("info message")')
                self.true(await stream.wait(6))
            with self.getAsyncLoggerStream(logname, 'warn message') as stream:
                await core.callStorm('$lib.log.warning("warn message")')
                self.true(await stream.wait(6))
            with self.getAsyncLoggerStream(logname, 'error message') as stream:
                await core.callStorm('$lib.log.error("error message")')
                self.true(await stream.wait(6))

            # Extra without structlog handler in place has no change in results
            with self.getAsyncLoggerStream(logname, 'debug message') as stream:
                await core.callStorm('$lib.log.debug("debug message", extra=({"key": "valu"}))')
                self.true(await stream.wait(6))
            stream.seek(0)
            self.eq(stream.read(), 'debug message\n')

            # Extra can be empty too
            with self.getAsyncLoggerStream(logname, 'debug message') as stream:
                await core.callStorm('$lib.log.debug("debug message", extra=({}))')
                self.true(await stream.wait(6))
            stream.seek(0)
            self.eq(stream.read(), 'debug message\n')

            # Extra must be a dict after toprim is called on him.
            with self.raises(s_exc.BadArg):
                await core.callStorm('$lib.log.debug("debug message", extra=(foo, bar, baz))')

            # structlog test
            with self.getStructuredAsyncLoggerStream(logname, '"key": "valu"') as stream:
                await core.callStorm('$lib.log.debug("struct1 message")')
                await core.callStorm('$lib.log.debug("struct2 message", extra=({"key": "valu"}))')
                self.true(await stream.wait(6))

            data = stream.getvalue()
            raw_mesgs = [m for m in data.split('\n') if m]
            msgs = [json.loads(m) for m in raw_mesgs]
            self.len(2, msgs)
            mesg = msgs[0]
            self.eq(mesg.get('logger').get('name'), 'synapse.storm.log')
            self.eq(mesg.get('message'), 'struct1 message')
            self.none(mesg.get('key'))

            mesg = msgs[1]
            self.eq(mesg.get('logger').get('name'), 'synapse.storm.log')
            self.eq(mesg.get('message'), 'struct2 message')
            self.eq(mesg.get('key'), 'valu')



import io
import json

import msgpack

import unittest.mock as mock

import synapse.lib.cell as s_cell
import synapse.lib.msgpack as s_msgpack
import synapse.tests.utils as s_t_utils
import synapse.tools.aha.list as s_a_list
import synapse.tools.aha.easycert as s_a_easycert

class AhaToolsTest(s_t_utils.SynTest):

    async def test_aha_list(self):

        ephemeral_address = 'tcp://0.0.0.0:0/'
        async with self.getTestAha(conf={'auth:passwd': 'root',
                                         'dmon:listen': ephemeral_address}) as aha:
            _, port = aha.sockaddr
            ahaurl = f'tcp://root:root@127.0.0.1:{port}'
            conf0 = {
                'aha:registry': ahaurl,
                'aha:name': 'cell0',
                'aha:network': 'demo.net',
                'dmon:listen': ephemeral_address,
            }
            conf1 = {
                'aha:registry': ahaurl,
                'aha:name': 'cell1',
                'aha:network': 'example.net',
                'dmon:listen': ephemeral_address,
            }
            waiter = aha.waiter(2, 'aha:svcadd')
            async with self.getTestCell(s_cell.Cell, conf=conf0) as cell0:
                async with self.getTestCell(s_cell.Cell, conf=conf1) as cell1:
                    self.true(await waiter.wait(timeout=6))
                    argv = [ahaurl]
                    retn, outp = await self.execToolMain(s_a_list._main, argv)
                    self.eq(retn, 0)
                    outp.expect('Service              network')
                    outp.expect('cell0                demo.net')
                    outp.expect('cell1                example.net')

                    argv = [ahaurl, 'demo.net']
                    retn, outp = await self.execToolMain(s_a_list._main, argv)
                    self.eq(retn, 0)
                    outp.expect('Service              network')
                    outp.expect('cell0                demo.net')
                    self.false(outp.expect('cell1                example.net', throw=False))

    async def test_aha_easycert(self):

        ephemeral_address = 'tcp://0.0.0.0:0/'
        async with self.getTestAha(conf={'auth:passwd': 'root',
                                         'dmon:listen': ephemeral_address}) as aha:
            _, port = aha.sockaddr
            ahaurl = f'tcp://root:root@127.0.0.1:{port}'

            with self.getTestDir() as dirn:
                argvbase = ['-a', ahaurl, '--certdir', dirn]
                argv = argvbase + ['--ca', 'demo.net']
                retn, outp = await self.execToolMain(s_a_easycert._main, argv)
                self.eq(retn, 0)
                outp.expect('Saved CA cert')
                outp.expect('cas/demo.net.crt')

                argv = argvbase + ['--server', '--server-sans', 'DNS:beeper.demo.net,DNS:booper.demo.net',
                                   '--network', 'demo.net', 'beep.demo.net']
                retn, outp = await self.execToolMain(s_a_easycert._main, argv)
                self.eq(retn, 0)
                outp.expect('key saved')
                outp.expect('hosts/beep.demo.net.key')
                outp.expect('crt saved')
                outp.expect('hosts/beep.demo.net.crt')

                argv = argvbase + ['--network', 'demo.net', 'mallory@demo.net']
                retn, outp = await self.execToolMain(s_a_easycert._main, argv)
                self.eq(retn, 0)
                outp.expect('key saved')
                outp.expect('users/mallory@demo.net.key')
                outp.expect('crt saved')
                outp.expect('users/mallory@demo.net.crt')



import asyncio
import logging
import datetime
from datetime import timezone as tz
from unittest import mock

import synapse.lib.cmdr as s_cmdr
import synapse.lib.provenance as s_provenance

import synapse.tests.utils as s_t_utils

MINSECS = 60
HOURSECS = 60 * MINSECS
DAYSECS = 24 * HOURSECS

logger = logging.getLogger(__name__)

class CmdCronTest(s_t_utils.SynTest):

    async def test_cron(self):
        MONO_DELT = 1543827303.0
        unixtime = datetime.datetime(year=2018, month=12, day=5, hour=7, minute=0, tzinfo=tz.utc).timestamp()
        s_provenance.reset()

        def timetime():
            return unixtime

        def looptime():
            return unixtime - MONO_DELT

        loop = asyncio.get_running_loop()

        with mock.patch.object(loop, 'time', looptime), mock.patch('time.time', timetime):
            async with self.getTestCoreAndProxy() as (realcore, core):

                outp = self.getTestOutp()
                async with await s_cmdr.getItemCmdr(core, outp=outp) as cmdr:

                    # Various silliness

                    await cmdr.runCmdLine('cron')
                    self.true(outp.expect('Manages cron jobs in a cortex'))
                    await cmdr.runCmdLine('cron timemachine')
                    self.true(outp.expect('invalid choice'))

                    await cmdr.runCmdLine('cron list')
                    self.true(outp.expect('No cron jobs found'))

                    await cmdr.runCmdLine('cron ls')
                    self.true(outp.expect('No cron jobs found'))

                    outp.clear()

                    await cmdr.runCmdLine("cron add -M+1,beeroclock {[graph:node='*' :type=m1]}")
                    self.true(outp.expect('failed to parse parameter'))

                    await cmdr.runCmdLine("cron add -m nosuchmonth -d=-2 {#foo}")
                    self.true(outp.expect('failed to parse fixed parameter'))

                    outp.clear()
                    await cmdr.runCmdLine("cron add -m 8nosuchmonth -d=-2 {#foo}")
                    self.true(outp.expect('failed to parse fixed parameter'))

                    await cmdr.runCmdLine("cron add -d Mon -m +3 {#foo}")
                    self.true(outp.expect('provide a recurrence value with day of week'))

                    await cmdr.runCmdLine("cron add -dMon -m June {#foo}")
                    self.true(outp.expect('fix month or year with day of week'))

                    await cmdr.runCmdLine("cron add -dMon -m +3 -y +2 {#foo}")
                    self.true(outp.expect('more than 1 recurrence'))

                    await cmdr.runCmdLine("cron add --year=2019 {#foo}")
                    self.true(outp.expect('year may not be a fixed value'))

                    await cmdr.runCmdLine("cron add {#foo}")
                    self.true(outp.expect('must provide at least one optional'))

                    await cmdr.runCmdLine("cron add -H3 -M +4 {#foo}")
                    self.true(outp.expect('fixed unit may not be larger'))

                    outp.clear()
                    await cmdr.runCmdLine('cron add -d Tuesday,1 {#foo}')
                    self.true(outp.expect('failed to parse day value'))

                    outp.clear()
                    await cmdr.runCmdLine('cron add -d Fri,3 {#foo}')
                    self.true(outp.expect('failed to parse day value'))

                    outp.clear()
                    await cmdr.runCmdLine('cron add }')
                    self.true(outp.expect('BadSyntax'))

                    # add a mechanism on which we can wait...
                    await realcore.nodes('$lib.queue.add(foo)')

                    async def getNextFoo():
                        return await asyncio.wait_for(realcore.callStorm('''
                            $foo = $lib.queue.get(foo)
                            ($offs, $retn) = $foo.get()
                            $foo.cull($offs)
                            return($retn)
                        '''), timeout=5)

                    ##################
                    # Start simple: add a cron job that creates a node every minute
                    outp.clear()
                    await cmdr.runCmdLine("cron add -M +1 {$lib.queue.get(foo).put(bar)}")
                    self.true(outp.expect('Created cron job'))
                    guid = outp.mesgs[-1].strip().rsplit(' ', 1)[-1]

                    unixtime += 60
                    self.eq('bar', await getNextFoo())

                    await cmdr.runCmdLine('cron list')
                    self.true(outp.expect('(bar)'))

                    # Make sure it ran
                    await cmdr.runCmdLine(f"cron mod {guid[:6]} {{$lib.queue.get(foo).put(baz)}}")
                    self.true(outp.expect('Modified cron job'))
                    await cmdr.runCmdLine(f"cron edit xxx {{[graph:node='*' :type=m2]}}")
                    self.true(outp.expect('does not match'))
                    await cmdr.runCmdLine(f"cron mod xxx yyy")
                    self.true(outp.expect('expected second argument to start with {'))

                    # Make sure the old one didn't run and the new query ran
                    unixtime += 60
                    self.eq('baz', await getNextFoo())

                    outp.clear()

                    # Delete the job
                    await cmdr.runCmdLine(f"cron del {guid}")
                    self.true(outp.expect('Deleted cron job'))
                    await cmdr.runCmdLine(f"cron del xxx")
                    self.true(outp.expect('does not match'))
                    await cmdr.runCmdLine(f"cron rm xxx")
                    self.true(outp.expect('does not match'))

                    # Make sure deleted job didn't run
                    unixtime += 60
                    await asyncio.sleep(0)
                    self.eq(0, await realcore.callStorm('return($lib.queue.get(foo).size())'))

                    # Test fixed minute, i.e. every hour at 17 past
                    unixtime = datetime.datetime(year=2018, month=12, day=5, hour=7, minute=10,
                                                 tzinfo=tz.utc).timestamp()
                    await cmdr.runCmdLine("cron add -M 17 {$lib.queue.get(foo).put(faz)}")
                    guid = outp.mesgs[-1].strip().rsplit(' ', 1)[-1]

                    unixtime += 7 * MINSECS

                    self.eq('faz', await getNextFoo())
                    await cmdr.runCmdLine(f"cron del {guid}")

                    ##################

                    # Test day increment
                    await cmdr.runCmdLine("cron add -d +2 {$lib.queue.get(foo).put(d1)}")
                    self.true(outp.expect('Created cron job'))
                    guid1 = outp.mesgs[-1].strip().rsplit(' ', 1)[-1]

                    unixtime += DAYSECS

                    # Make sure it *didn't* run
                    await asyncio.sleep(0)
                    self.eq(0, await realcore.callStorm('return($lib.queue.get(foo).size())'))

                    unixtime += DAYSECS

                    self.eq('d1', await getNextFoo())

                    unixtime += DAYSECS * 2

                    outp.clear()
                    self.eq('d1', await getNextFoo())
                    await cmdr.runCmdLine(f"cron del {guid1}")
                    outp.expect('Deleted cron job')

                    ##################

                    # Test fixed day of week: every Monday and Thursday at 3am
                    unixtime = datetime.datetime(year=2018, month=12, day=11, hour=7, minute=10,
                                                 tzinfo=tz.utc).timestamp()  # A Tuesday

                    outp.clear()
                    await cmdr.runCmdLine("cron add -H 3 -d Mon,Thursday {$lib.queue.get(foo).put(d2)}")
                    self.true(outp.expect('Created cron job'))

                    guid2 = outp.mesgs[-1].strip().rsplit(' ', 1)[-1]
                    unixtime = datetime.datetime(year=2018, month=12, day=12, hour=3, minute=10,
                                                 tzinfo=tz.utc).timestamp()  # Now Wednesday

                    outp.clear()
                    await cmdr.runCmdLine(f'cron stat {guid2}')
                    self.true(outp.expect('last start time: Never'))

                    unixtime = datetime.datetime(year=2018, month=12, day=13, hour=3, minute=10,
                                                 tzinfo=tz.utc).timestamp()  # Now Thursday

                    self.eq('d2', await getNextFoo())

                    outp.clear()
                    await cmdr.runCmdLine(f'cron stat {guid2}')
                    self.true(outp.expect('last start time: 2018'))
                    self.true(outp.expect('dayofweek       0'))

                    outp.clear()
                    await cmdr.runCmdLine(f"cron del {guid2}")
                    outp.expect('Deleted cron job')

                    await cmdr.runCmdLine("cron add -H 3 -d Noday {[graph:node='*' :type=d2]}")
                    self.true(outp.expect('failed to parse day value "Noday"'))

                    ##################

                    # Test fixed day of month: second-to-last day of month
                    await cmdr.runCmdLine("cron add -d-2 -mDec {$lib.queue.get(foo).put(d3)}")
                    guid = outp.mesgs[-1].strip().rsplit(' ', 1)[-1]

                    unixtime = datetime.datetime(year=2018, month=12, day=29, hour=0, minute=0,
                                                 tzinfo=tz.utc).timestamp()  # Now Thursday

                    unixtime += DAYSECS

                    self.eq('d3', await getNextFoo())

                    outp.clear()
                    await cmdr.runCmdLine(f"cron del {guid}")
                    outp.expect('Deleted cron job')

                    ##################

                    # Test month increment

                    outp.clear()
                    await cmdr.runCmdLine("cron add -m +2 -d=4 {$lib.queue.get(foo).put(month1)}")
                    guid = outp.mesgs[-1].strip().rsplit(' ', 1)[-1]
                    self.true(outp.expect('Created cron job'))

                    unixtime = datetime.datetime(year=2019, month=2, day=4, hour=0, minute=0,
                                                 tzinfo=tz.utc).timestamp()  # Now Thursday

                    self.eq('month1', await getNextFoo())

                    outp.clear()
                    await cmdr.runCmdLine(f"cron del {guid}")
                    outp.expect('Deleted cron job')

                    ##################

                    # Test year increment

                    outp.clear()
                    await cmdr.runCmdLine("cron add -y +2 {$lib.queue.get(foo).put(year1)}")
                    guid2 = outp.mesgs[-1].strip().rsplit(' ', 1)[-1]
                    self.true(outp.expect('Created cron job'))

                    unixtime = datetime.datetime(year=2021, month=1, day=1, hour=0, minute=0,
                                                 tzinfo=tz.utc).timestamp()  # Now Thursday
                    self.eq('year1', await getNextFoo())

                    outp.clear()
                    await cmdr.runCmdLine(f'cron stat {guid2[:6]}')
                    self.true(outp.expect("{'month': 1, 'hour': 0, 'minute': 0, 'dayofmonth': 1}"))

                    outp.clear()
                    await cmdr.runCmdLine(f"cron del {guid2}")
                    outp.expect('Deleted cron job')

                    # Make sure second-to-last day works for February
                    outp.clear()
                    await cmdr.runCmdLine("cron add -m February -d=-2 {$lib.queue.get(foo).put(year2)}")
                    self.true(outp.expect('Created cron job'))

                    unixtime = datetime.datetime(year=2021, month=2, day=27, hour=0, minute=0,
                                                 tzinfo=tz.utc).timestamp()  # Now Thursday

                    self.eq('year2', await getNextFoo())

                    ##################

                    # Test 'at' command
                    outp.clear()
                    await cmdr.runCmdLine('at')
                    self.true(outp.expect('Adds a non-recurring'))

                    await cmdr.runCmdLine('at --not-a-real-flag')
                    self.true(outp.expect('the following arguments'))

                    await cmdr.runCmdLine('at {#foo} {#bar}')
                    self.true(outp.expect('only a single query'))

                    await cmdr.runCmdLine('at {#foo}')
                    self.true(outp.expect('at least'))

                    await cmdr.runCmdLine('at +1')
                    self.true(outp.expect('missing unit'))

                    await cmdr.runCmdLine('at +1parsec')
                    self.true(outp.expect('Trouble parsing'))

                    await cmdr.runCmdLine('at +1day')
                    self.true(outp.expect('Missing query'))

                    await cmdr.runCmdLine("at +5 minutes {$lib.queue.get(foo).put(at1)}")

                    unixtime += 5 * MINSECS

                    self.eq('at1', await getNextFoo())

                    await cmdr.runCmdLine("at +1 day +7 days {$lib.queue.get(foo).put(at2)}")
                    guid = outp.mesgs[-1].strip().rsplit(' ', 1)[-1]

                    unixtime += DAYSECS

                    self.eq('at2', await getNextFoo())

                    unixtime += 6 * DAYSECS + 1

                    self.eq('at2', await getNextFoo())

                    await cmdr.runCmdLine("at 202104170415 {$lib.queue.get(foo).put(at3)}")

                    unixtime = datetime.datetime(year=2021, month=4, day=17, hour=4, minute=15,
                                                 tzinfo=tz.utc).timestamp()  # Now Thursday

                    self.eq('at3', await getNextFoo())
                    outp.clear()

                    ##################

                    # Test 'stat' command
                    await cmdr.runCmdLine(f'cron stat xxx')
                    self.true(outp.expect('provided iden does not match any'))

                    await cmdr.runCmdLine(f'cron stat {guid[:6]}')
                    self.true(outp.expect('last result:     finished successfully with 0 nodes'))
                    self.true(outp.expect('entries:         <None>'))

                    ##################

                    # Test 'enable' 'disable' commands
                    await cmdr.runCmdLine(f'cron enable xxx')
                    self.true(outp.expect('provided iden does not match any'))
                    outp.clear()

                    await cmdr.runCmdLine(f'cron disable xxx')
                    self.true(outp.expect('provided iden does not match any'))
                    outp.clear()

                    await cmdr.runCmdLine(f'cron disable {guid[:6]}')
                    await cmdr.runCmdLine(f'cron stat {guid[:6]}')
                    self.true(outp.expect(f'enabled:         N'))
                    outp.clear()
                    await cmdr.runCmdLine(f'cron enable {guid[:6]}')
                    await cmdr.runCmdLine(f'cron stat {guid[:6]}')
                    self.true(outp.expect(f'enabled:         Y'))
                    outp.clear()

                    ###################

                    # Delete an expired at job
                    outp.clear()
                    await cmdr.runCmdLine(f"cron del {guid}")
                    self.true(outp.expect('Deleted cron job'))

                    ##################

                    # Test the aliases
                    outp.clear()
                    await cmdr.runCmdLine('cron add --hourly 15 {#bar}')
                    guid = outp.mesgs[-1].strip().rsplit(' ', 1)[-1]
                    await cmdr.runCmdLine(f'cron stat {guid[:6]}')
                    self.true(outp.expect("{'minute': 15}"))

                    outp.clear()
                    await cmdr.runCmdLine('cron add --daily 05:47 {#bar}')
                    guid = outp.mesgs[-1].strip().rsplit(' ', 1)[-1]
                    await cmdr.runCmdLine(f'cron stat {guid[:6]}')
                    self.true(outp.expect("{'hour': 5, 'minute': 47"))

                    outp.clear()
                    await cmdr.runCmdLine('cron add --monthly=-1:12:30 {#bar}')
                    guid = outp.mesgs[-1].strip().rsplit(' ', 1)[-1]
                    await cmdr.runCmdLine(f'cron stat {guid[:6]}')
                    self.true(outp.expect("{'hour': 12, 'minute': 30, 'dayofmonth': -1}"))

                    outp.clear()
                    await cmdr.runCmdLine('cron add --yearly 04:17:12:30 {#bar}')
                    guid = outp.mesgs[-1].strip().rsplit(' ', 1)[-1]
                    await cmdr.runCmdLine(f'cron stat {guid[:6]}')
                    self.true(outp.expect("{'month': 4, 'hour': 12, 'minute': 30, 'dayofmonth': 17}"))

                    outp.clear()
                    await cmdr.runCmdLine('cron add --yearly 04:17:12 {#bar}')
                    self.true(outp.expect('Failed to parse parameter'))

                    outp.clear()
                    await cmdr.runCmdLine('cron add --daily xx:xx {#bar}')
                    self.true(outp.expect('Failed to parse ..ly parameter'))

                    outp.clear()
                    await cmdr.runCmdLine('cron add --hourly 1 -M 17 {#bar}')
                    self.true(outp.expect('may not use both'))

                # Test manipulating cron jobs as another user
                bond = await realcore.auth.addUser('bond')

                async with realcore.getLocalProxy(user='bond') as tcore:
                    toutp = self.getTestOutp()
                    tcmdr = await s_cmdr.getItemCmdr(tcore, outp=toutp)

                    await tcmdr.runCmdLine('cron list')
                    self.true(toutp.expect('No cron jobs found'))

                    toutp.clear()
                    await tcmdr.runCmdLine(f'cron disable {guid[:6]}')
                    self.true(toutp.expect('provided iden does not match'))

                    toutp.clear()
                    await tcmdr.runCmdLine(f'cron enable {guid[:6]}')
                    self.true(toutp.expect('provided iden does not match'))

                    toutp.clear()
                    await tcmdr.runCmdLine(f'cron edit {guid[:6]} {{#foo}}')
                    self.true(toutp.expect('provided iden does not match'))

                    toutp.clear()
                    await tcmdr.runCmdLine(f'cron del {guid[:6]}')
                    self.true(toutp.expect('provided iden does not match'))

                    # Give explicit perm
                    await core.addUserRule(bond.iden, (True, ('cron', 'get')))

                    toutp.clear()
                    await tcmdr.runCmdLine('cron list')
                    self.true(toutp.expect('root'))

                    await core.addUserRule(bond.iden, (True, ('cron', 'set')))

                    toutp.clear()
                    await tcmdr.runCmdLine(f'cron disable {guid[:6]}')
                    self.true(toutp.expect('Disabled cron job'))

                    toutp.clear()
                    await tcmdr.runCmdLine(f'cron enable {guid[:6]}')
                    self.true(toutp.expect('Enabled cron job'))

                    toutp.clear()
                    await tcmdr.runCmdLine(f'cron edit {guid[:6]} {{#foo}}')
                    self.true(toutp.expect('Modified cron job'))

                    await core.addUserRule(bond.iden, (True, ('cron', 'del')))

                    toutp.clear()
                    await tcmdr.runCmdLine(f'cron del {guid[:6]}')
                    self.true(toutp.expect('Deleted cron job'))



import synapse.exc as s_exc

import synapse.lib.types as s_types
import synapse.lib.module as s_module

class DnsName(s_types.Str):

    def postTypeInit(self):

        s_types.Str.postTypeInit(self)
        self.inarpa = '.in-addr.arpa'
        self.inarpa6 = '.ip6.arpa'

        self.setNormFunc(str, self._normPyStr)

    def _normPyStr(self, valu):
        # Backwards compatible
        norm = valu.lower()
        norm = norm.strip()  # type: str
        # Break out fqdn / ipv4 / ipv6 subs :D
        subs = {}
        # ipv4
        if norm.isnumeric():
            # do-nothing for integer only strs
            pass
        elif norm.endswith(self.inarpa):
            # Strip, reverse, check if ipv4
            temp = norm[:-len(self.inarpa)]
            temp = '.'.join(temp.split('.')[::-1])
            try:
                ipv4norm, info = self.modl.type('inet:ipv4').norm(temp)
            except s_exc.BadTypeValu as e:
                pass
            else:
                subs['ipv4'] = ipv4norm
        elif norm.endswith(self.inarpa6):
            parts = [c for c in norm[:-len(self.inarpa6)][::-1] if c != '.']
            try:
                if len(parts) != 32:
                    raise s_exc.BadTypeValu(mesg='Invalid number of ipv6 parts')
                temp = int(''.join(parts), 16)
                ipv6norm, info = self.modl.type('inet:ipv6').norm(temp)
            except s_exc.BadTypeValu as e:
                pass
            else:
                subs['ipv6'] = ipv6norm
                ipv4 = info.get('subs').get('ipv4')
                if ipv4 is not None:
                    subs['ipv4'] = ipv4
        else:
            # Try fallbacks to parse out possible ipv4/ipv6 garbage queries
            try:
                ipv4norm, info = self.modl.type('inet:ipv4').norm(norm)
            except s_exc.BadTypeValu as e:
                try:
                    ipv6norm, info = self.modl.type('inet:ipv6').norm(norm)
                except s_exc.BadTypeValu as e2:
                    pass
                else:
                    subs['ipv6'] = ipv6norm
                    ipv4 = info.get('subs').get('ipv4')
                    if ipv4 is not None:
                        subs['ipv4'] = ipv4
            else:
                subs['ipv4'] = ipv4norm

            # Lastly, try give the norm'd valu a shot as an inet:fqdn
            try:
                fqdnnorm, info = self.modl.type('inet:fqdn').norm(norm)
            except s_exc.BadTypeValu as e:
                pass
            else:
                subs['fqdn'] = fqdnnorm

        return norm, {'subs': subs}

class DnsModule(s_module.CoreModule):

    def getModelDefs(self):

        modl = {

            'ctors': (

                ('inet:dns:name', 'synapse.models.dns.DnsName', {}, {
                    'doc': 'A DNS query name string. Likely an FQDN but not always.',
                    'ex': 'vertex.link',
                }),

            ),

            'types': (

                ('inet:dns:a', ('comp', {'fields': (('fqdn', 'inet:fqdn'), ('ipv4', 'inet:ipv4'))}), {
                    'ex': '(vertex.link,1.2.3.4)',
                    'doc': 'The result of a DNS A record lookup.'}),

                ('inet:dns:aaaa', ('comp', {'fields': (('fqdn', 'inet:fqdn'), ('ipv6', 'inet:ipv6'))}), {
                    'ex': '(vertex.link,2607:f8b0:4004:809::200e)',
                    'doc': 'The result of a DNS AAAA record lookup.'}),

                ('inet:dns:rev', ('comp', {'fields': (('ipv4', 'inet:ipv4'), ('fqdn', 'inet:fqdn'))}), {
                    'ex': '(1.2.3.4,vertex.link)',
                    'doc': 'The transformed result of a DNS PTR record lookup.'}),

                ('inet:dns:rev6', ('comp', {'fields': (('ipv6', 'inet:ipv6'), ('fqdn', 'inet:fqdn'))}), {
                    'ex': '(2607:f8b0:4004:809::200e,vertex.link)',
                    'doc': 'The transformed result of a DNS PTR record for an IPv6 address.'}),

                ('inet:dns:ns', ('comp', {'fields': (('zone', 'inet:fqdn'), ('ns', 'inet:fqdn'))}), {
                    'ex': '(vertex.link,ns.dnshost.com)',
                    'doc': 'The result of a DNS NS record lookup.'}),

                ('inet:dns:cname', ('comp', {'fields': (('fqdn', 'inet:fqdn'), ('cname', 'inet:fqdn'))}), {
                    'ex': '(foo.vertex.link,vertex.link)',
                    'doc': 'The result of a DNS CNAME record lookup.'}),

                ('inet:dns:mx', ('comp', {'fields': (('fqdn', 'inet:fqdn'), ('mx', 'inet:fqdn'))}), {
                    'ex': '(vertex.link,mail.vertex.link)',
                    'doc': 'The result of a DNS MX record lookup.'}),

                ('inet:dns:soa', ('guid', {}), {
                    'doc': 'The result of a DNS SOA record lookup.'}),

                ('inet:dns:txt', ('comp', {'fields': (('fqdn', 'inet:fqdn'), ('txt', 'str'))}), {
                    'ex': '(hehe.vertex.link,"fancy TXT record")',
                    'doc': 'The result of a DNS MX record lookup.'}),

                ('inet:dns:type', ('int', {}), {
                    'doc': 'A DNS query/answer type integer.'}),

                ('inet:dns:query',
                    ('comp', {'fields': (('client', 'inet:client'), ('name', 'inet:dns:name'), ('type', 'int'))}), {
                        'ex': '(1.2.3.4, woot.com, 1)',
                        'doc': 'A DNS query unique to a given client.'}),

                ('inet:dns:request', ('guid', {}), {
                    'doc': 'A single instance of a DNS resolver request and optional reply info.'}),

                ('inet:dns:answer', ('guid', {}), {
                    'doc': 'A single answer from within a DNS reply.'}),

                ('inet:dns:wild:a', ('comp', {'fields': (('fqdn', 'inet:fqdn'), ('ipv4', 'inet:ipv4'))}), {
                    'doc': 'A DNS A wild card record and the IPv4 it resolves to.'}),

                ('inet:dns:wild:aaaa', ('comp', {'fields': (('fqdn', 'inet:fqdn'), ('ipv6', 'inet:ipv6'))}), {
                    'doc': 'A DNS AAAA wild card record and the IPv6 it resolves to.'}),
            ),

            'forms': (
                ('inet:dns:a', {}, (
                    ('fqdn', ('inet:fqdn', {}), {'ro': True,
                        'doc': 'The domain queried for its DNS A record.'}),
                    ('ipv4', ('inet:ipv4', {}), {'ro': True,
                        'doc': 'The IPv4 address returned in the A record.'}),
                )),
                ('inet:dns:aaaa', {}, (
                    ('fqdn', ('inet:fqdn', {}), {'ro': True,
                         'doc': 'The domain queried for its DNS AAAA record.'}),
                    ('ipv6', ('inet:ipv6', {}), {'ro': True,
                         'doc': 'The IPv6 address returned in the AAAA record.'}),
                )),
                ('inet:dns:rev', {}, (
                    ('ipv4', ('inet:ipv4', {}), {'ro': True,
                         'doc': 'The IPv4 address queried for its DNS PTR record.'}),
                    ('fqdn', ('inet:fqdn', {}), {'ro': True,
                         'doc': 'The domain returned in the PTR record.'}),
                )),
                ('inet:dns:rev6', {}, (
                    ('ipv6', ('inet:ipv6', {}), {'ro': True,
                         'doc': 'The IPv6 address queried for its DNS PTR record.'}),
                    ('fqdn', ('inet:fqdn', {}), {'ro': True,
                         'doc': 'The domain returned in the PTR record.'}),
                )),
                ('inet:dns:ns', {}, (
                    ('zone', ('inet:fqdn', {}), {'ro': True,
                         'doc': 'The domain queried for its DNS NS record.'}),
                    ('ns', ('inet:fqdn', {}), {'ro': True,
                         'doc': 'The domain returned in the NS record.'}),
                )),
                ('inet:dns:cname', {}, (
                    ('fqdn', ('inet:fqdn', {}), {'ro': True,
                         'doc': 'The domain queried for its CNAME record.'}),
                    ('cname', ('inet:fqdn', {}), {'ro': True,
                         'doc': 'The domain returned in the CNAME record.'}),
                )),
                ('inet:dns:mx', {}, (
                    ('fqdn', ('inet:fqdn', {}), {'ro': True,
                         'doc': 'The domain queried for its MX record.'}),
                    ('mx', ('inet:fqdn', {}), {'ro': True,
                         'doc': 'The domain returned in the MX record.'}),
                )),

                ('inet:dns:soa', {}, (
                    ('fqdn', ('inet:fqdn', {}), {'ro': True,
                         'doc': 'The domain queried for its SOA record.'}),
                    ('ns', ('inet:fqdn', {}), {'ro': True,
                         'doc': 'The domain (MNAME) returned in the SOA record.'}),
                    ('email', ('inet:email', {}), {'ro': True,
                         'doc': 'The email address (RNAME) returned in the SOA record.'}),
                )),

                ('inet:dns:txt', {}, (
                    ('fqdn', ('inet:fqdn', {}), {'ro': True,
                         'doc': 'The domain queried for its TXT record.'}),
                    ('txt', ('str', {}), {'ro': True,
                         'doc': 'The string returned in the TXT record.'}),
                )),

                ('inet:dns:query', {}, (
                    ('client', ('inet:client', {}), {'ro': True, }),
                    ('name', ('inet:dns:name', {}), {'ro': True, }),
                    ('name:ipv4', ('inet:ipv4', {}), {}),
                    ('name:ipv6', ('inet:ipv6', {}), {}),
                    ('name:fqdn', ('inet:fqdn', {}), {}),
                    ('type', ('int', {}), {'ro': True, }),
                )),

                ('inet:dns:request', {}, (

                    ('time', ('time', {}), {}),

                    ('query', ('inet:dns:query', {}), {}),
                    ('query:name', ('inet:dns:name', {}), {}),
                    ('query:name:ipv4', ('inet:ipv4', {}), {}),
                    ('query:name:ipv6', ('inet:ipv6', {}), {}),
                    ('query:name:fqdn', ('inet:fqdn', {}), {}),
                    ('query:type', ('int', {}), {}),

                    ('server', ('inet:server', {}), {}),

                    ('reply:code', ('int', {}), {
                        'doc': 'The DNS server response code.'}),

                    ('exe', ('file:bytes', {}), {
                        'doc': 'The file containing the code that attempted the DNS lookup.'}),

                    ('proc', ('it:exec:proc', {}), {
                        'doc': 'The process that attempted the DNS lookup.'}),

                    ('host', ('it:host', {}), {
                        'doc': 'The host that attempted the DNS lookup.'}),

                )),

                ('inet:dns:answer', {}, (

                    ('ttl', ('int', {}), {}),
                    ('request', ('inet:dns:request', {}), {}),

                    ('a', ('inet:dns:a', {}), {'ro': True,
                        'doc': 'The DNS A record returned by the lookup.',
                    }),
                    ('ns', ('inet:dns:ns', {}), {'ro': True,
                        'doc': 'The DNS NS record returned by the lookup.',
                    }),
                    ('rev', ('inet:dns:rev', {}), {'ro': True,
                        'doc': 'The DNS PTR record returned by the lookup.',
                    }),
                    ('aaaa', ('inet:dns:aaaa', {}), {'ro': True,
                        'doc': 'The DNS AAAA record returned by the lookup.',
                    }),
                    ('rev6', ('inet:dns:rev6', {}), {'ro': True,
                        'doc': 'The DNS PTR record returned by the lookup of an IPv6 address.',
                    }),
                    ('cname', ('inet:dns:cname', {}), {'ro': True,
                        'doc': 'The DNS CNAME record returned by the lookup.',
                    }),
                    ('mx', ('inet:dns:mx', {}), {'ro': True,
                        'doc': 'The DNS MX record returned by the lookup.',
                    }),
                    ('soa', ('inet:dns:soa', {}), {'ro': True,
                        'doc': 'The domain queried for its SOA record.',
                    }),
                    ('txt', ('inet:dns:txt', {}), {'ro': True,
                        'doc': 'The DNS TXT record returned by the lookup.',
                    }),
                )),

                ('inet:dns:wild:a', {}, (
                    ('fqdn', ('inet:fqdn', {}), {'ro': True,
                        'doc': 'The domain containing a wild card record.'}),
                    ('ipv4', ('inet:ipv4', {}), {'ro': True,
                        'doc': 'The IPv4 address returned by wild card resolutions.'}),
                )),

                ('inet:dns:wild:aaaa', {}, (
                    ('fqdn', ('inet:fqdn', {}), {'ro': True,
                        'doc': 'The domain containing a wild card record.'}),
                    ('ipv6', ('inet:ipv6', {}), {'ro': True,
                        'doc': 'The IPv6 address returned by wild card resolutions.'}),
                )),
            )

        }
        name = 'inet:dns'
        return ((name, modl), )



import shlex
import synapse.exc as s_exc

import synapse.lib.cmd as s_cmd
import synapse.lib.cli as s_cli
import synapse.lib.time as s_time

class PsCmd(s_cli.Cmd):

    '''
    List running tasks in the cortex.
    '''

    _cmd_name = 'ps'
    _cmd_syntax = (
        ('--verbose', {}),
        ('-v', {}),
    )

    async def runCmdOpts(self, opts):

        core = self.getCmdItem()
        tasks = await core.ps()
        isverbose = opts.get('verbose') or opts.get('v')
        MAXFIELDLEN = 120

        def clamp(field):
            if isinstance(field, dict):
                for key, valu in field.items():
                    field[key] = clamp(valu)
            elif isinstance(field, str) and len(field) > MAXFIELDLEN:
                field = field[:MAXFIELDLEN] + '...'
            return field

        for task in tasks:

            self.printf('task iden: %s' % (task.get('iden'),))
            self.printf('    name: %s' % (task.get('name'),))
            self.printf('    user: %r' % (task.get('user'),))
            self.printf('    status: %r' % (task.get('status'),))
            metadata = task.get('info')
            if metadata is not None and not isverbose:
                metadata = clamp(metadata)

            self.printf('    metadata: %r' % metadata)
            self.printf('    start time: %s' % (s_time.repr(task.get('tick', 0)),))

        self.printf('%d tasks found.' % (len(tasks,)))

class KillCmd(s_cli.Cmd):
    '''
    Kill a running task/query within the cortex.

    Syntax:
        kill <iden>

    Users may specify a partial iden GUID in order to kill
    exactly one matching process based on the partial guid.
    '''
    _cmd_name = 'kill'
    _cmd_syntax = (
        ('line', {'type': 'glob'}),
    )

    def _make_argparser(self):
        parser = s_cmd.Parser(prog='kill', outp=self, description=self.__doc__)
        parser.add_argument('iden', help='Task iden to kill.', type=str)
        return parser

    async def runCmdOpts(self, opts):

        line = opts.get('line')
        if line is None:
            self.printf(self.__doc__)
            return

        try:
            opts = self._make_argparser().parse_args(shlex.split(line))
        except s_exc.ParserExit:
            return

        core = self.getCmdItem()

        match = opts.iden
        idens = []
        for task in await core.ps():
            iden = task.get('iden')
            if iden.startswith(match):
                idens.append(iden)

        if len(idens) == 0:
            self.printf('no matching process found.')
            return

        if len(idens) > 1:  # pragma: no cover
            # this is a non-trivial situation to test since the
            # boss idens are random guids
            self.printf('multiple matching processes found. aborting.')
            return

        kild = await core.kill(idens[0])
        self.printf('kill status: %r' % (kild,))



import unittest.mock as mock

import synapse.common as s_common

import synapse.tests.utils as s_t_utils

import synapse.tools.cellauth as s_cellauth

class CellAuthTest(s_t_utils.SynTest):

    async def test_cellauth_bad(self):

        async with self.getTestCore() as core:

            coreurl = core.getLocalUrl()

            argv = [coreurl]
            outp = self.getTestOutp()
            self.eq(await s_cellauth.main(argv, outp), -1)
            outp.expect('the following arguments are required:')

            outp.clear()
            argv = [coreurl, 'modify', '--adduser', 'foo', '--object', 'foo:bar']
            await s_cellauth.main(argv, outp)
            outp.expect('only valid with --addrule')

            def fakevers(self):
                return (0, 0, 0)

            with mock.patch('synapse.telepath.Proxy._getSynVers', fakevers):
                argv = [coreurl, 'modify', '--adduser', 'foo']
                outp = self.getTestOutp()
                await s_cellauth.main(argv, outp)
                outp.expect('Cell version 0.0.0 is outside of the cellauth supported range')

                argv = [coreurl, 'list']
                outp = self.getTestOutp()
                await s_cellauth.main(argv, outp)
                outp.expect('Cell version 0.0.0 is outside of the cellauth supported range')

    async def test_cellauth_list(self):

        async with self.getTestCore() as core:

            await self.addCreatorDeleterRoles(core)

            coreurl = core.getLocalUrl()

            argv = [coreurl, 'list']
            outp = self.getTestOutp()
            self.eq(await s_cellauth.main(argv, outp), 0)
            outp.expect('getting users and roles')
            outp.expect('users:')
            outp.expect('root')
            outp.expect('roles:')

            argv = [coreurl, '--debug', 'list', 'icanadd']
            outp = self.getTestOutp()
            self.eq(await s_cellauth.main(argv, outp), 0)

            outp.expect('icanadd')
            outp.expect('admin: False')
            outp.expect('role: creator')
            self.false(outp.expect('allow: node.add', throw=False))

            argv = [coreurl, 'list', 'creator']
            outp = self.getTestOutp()
            self.eq(await s_cellauth.main(argv, outp), 0)
            outp.expect('creator')
            outp.expect('type: role')

            argv = [coreurl, 'list', '--detail', 'icanadd']
            outp = self.getTestOutp()
            self.eq(await s_cellauth.main(argv, outp), 0)
            outp.expect('icanadd')
            outp.expect('admin: False')
            outp.expect('role: creator')
            outp.expect('allow: node.add', throw=False)

    async def test_cellauth_user(self):

        async with self.getTestCore() as core:

            coreurl = core.getLocalUrl()

            outp = self.getTestOutp()
            argv = [coreurl, 'modify', 'root']
            await s_cellauth.main(argv, outp)
            outp.expect('type: user')
            outp.expect('admin: True')
            outp.expect('locked: False')
            outp.expect('rules:')
            outp.expect('roles:')

            outp = self.getTestOutp()
            argv = [coreurl, 'modify', 'foo']
            await s_cellauth.main(argv, outp)
            outp.expect(f'no such user: foo')

            outp = self.getTestOutp()
            argv = [coreurl, 'modify', '--adduser', 'foo']
            await s_cellauth.main(argv, outp)
            outp.expect('adding user: foo')

            outp = self.getTestOutp()
            argv = [coreurl, 'modify', '--addrole', 'frole']
            await s_cellauth.main(argv, outp)
            outp.expect('adding role: frole')

            outp = self.getTestOutp()
            argv = [coreurl, 'modify', '--admin', 'foo']
            await s_cellauth.main(argv, outp)
            outp.expect('admin: True')

            outp = self.getTestOutp()
            argv = [coreurl, 'modify', '--noadmin', 'foo']
            await s_cellauth.main(argv, outp)
            outp.expect('admin: False')

            outp = self.getTestOutp()
            argv = [coreurl, 'modify', '--delrole', 'frole']
            await s_cellauth.main(argv, outp)
            outp.expect('deleting role: frole')

            outp = self.getTestOutp()
            argv = [coreurl, 'modify', '--deluser', 'foo']
            await s_cellauth.main(argv, outp)
            outp.expect('deleting user: foo')

    async def test_cellauth_lock(self):

        async with self.getTestCore() as core:

            coreurl = core.getLocalUrl()

            outp = self.getTestOutp()
            argv = [coreurl, 'modify', '--adduser', 'foo']
            await s_cellauth.main(argv, outp)

            outp = self.getTestOutp()
            argv = [coreurl, 'modify', '--lock', 'foo']
            await s_cellauth.main(argv, outp)
            outp.expect('locking user: foo')
            outp.expect('locked: True')

            outp = self.getTestOutp()
            argv = [coreurl, 'modify', '--unlock', 'foo']
            await s_cellauth.main(argv, outp)
            outp.expect('unlocking user: foo')
            outp.expect('locked: False')

    async def test_cellauth_passwd(self):

        with self.getTestSynDir() as syndir:

            s_common.yamlsave({'version': 1}, syndir, 'telepath.yaml')

            async with self.getTestCore() as core:

                coreurl = core.getLocalUrl()

                outp = self.getTestOutp()
                argv = [coreurl, 'modify', '--adduser', 'foo']
                await s_cellauth.main(argv, outp)

                outp = self.getTestOutp()
                argv = [coreurl, 'modify', '--passwd', 'mysecret', 'foo']
                await s_cellauth.main(argv, outp)
                outp.expect('setting passwd for: foo')

    async def test_cellauth_grants(self):

        async with self.getTestCore() as core:

            coreurl = core.getLocalUrl()

            outp = self.getTestOutp()
            argv = [coreurl, 'modify', '--adduser', 'foo']
            await s_cellauth.main(argv, outp)

            outp = self.getTestOutp()
            argv = [coreurl, 'modify', '--addrole', 'bar']
            await s_cellauth.main(argv, outp)

            argv = [coreurl, 'modify', '--grant', 'bar', 'foo']
            await s_cellauth.main(argv, outp)
            outp.expect('granting bar to: foo')
            outp.expect('role: bar')

            argv = [coreurl, 'modify', '--revoke', 'bar', 'foo']
            await s_cellauth.main(argv, outp)
            outp.expect('revoking bar from: foo')

            argv = [coreurl, 'modify', 'foo', '--setroles', 'bar', 'all']
            r = await s_cellauth.main(argv, outp)
            self.eq(0, r)
            outp.expect("settings roles ['bar', 'all'] to: foo")

    async def test_cellauth_rules(self):

        async with self.getTestCoreAndProxy() as (core, prox):

            coreurl = core.getLocalUrl()

            rule = 'node.add'
            rulerepr = repr((True, ['node', 'add']))
            nrule = '!node.add'
            nrulerepr = repr((False, ['node', 'add']))
            name = 'foo'

            outp = self.getTestOutp()
            argv = ['--debug', coreurl, 'modify', '--adduser', name]
            await s_cellauth.main(argv, outp)

            outp.clear()
            argv = [coreurl, 'modify', '--addrule', rule, name]
            await s_cellauth.main(argv, outp)
            outp.expect(f'adding rule to {name}: {rulerepr}')
            outp.expect(f'0 allow: node.add')

            user = await prox.getUserInfo(name)
            self.eq(user.get('rules'),
                    ((True, ('node', 'add',)),))

            outp.clear()
            argv = [coreurl, 'modify', '--addrule', 'node.del', 'foo']
            await s_cellauth.main(argv, outp)
            outp.expect(f'1 allow: node.del')

            outp.clear()
            argv = [coreurl, 'modify', '--addrule', 'service.get', 'foo']
            await s_cellauth.main(argv, outp)
            outp.expect(f'2 allow: service.get')

            outp.clear()
            argv = [coreurl, 'list', 'foo']
            await s_cellauth.main(argv, outp)
            outp.expect(f'0 allow: node.add')
            outp.expect(f'1 allow: node.del')
            outp.expect(f'2 allow: service.get')

            outp.clear()
            argv = [coreurl, 'modify', '--delrule', '1', 'foo']
            await s_cellauth.main(argv, outp)
            outp.expect(f'deleting rule index: 1')
            outp.expect(f'0 allow: node.add')
            outp.expect(f'1 allow: service.get')
            self.false(outp.expect(f'allow: node.del', throw=False))

            user = await prox.getUserInfo(name)
            self.eq(user.get('rules'),
                    ((True, ('node', 'add')), (True, ('service', 'get'))))

            outp.clear()
            argv = [coreurl, 'list', 'foo']
            await s_cellauth.main(argv, outp)
            outp.expect(f'0 allow: node.add')
            outp.expect(f'1 allow: service.get')
            self.false(outp.expect(f'allow: node.del', throw=False))

            outp.clear()
            argv = [coreurl, 'modify', '--delrule', '0', 'foo']
            await s_cellauth.main(argv, outp)
            outp.expect(f'deleting rule index: 0')
            outp.expect(f'0 allow: service.get')
            self.false(outp.expect(f'allow: node.add', throw=False))

            outp.clear()
            argv = [coreurl, 'modify', '--delrule', '0', 'foo']
            await s_cellauth.main(argv, outp)
            outp.expect(f'deleting rule index: 0')
            self.false(outp.expect(f'allow: service.get', throw=False))

            user = await prox.getUserInfo(name)
            self.eq(user.get('rules'), ())

            outp.clear()
            viewiden = core.view.iden
            argv = [coreurl, 'modify', '--addrule', nrule, name, '--object', viewiden]
            await s_cellauth.main(argv, outp)
            outp.expect(f'adding rule to {name}: {nrulerepr}')
            outp.expect(f'0 deny: node.add')

            outp.clear()
            argv = [coreurl, 'modify', '--addrule', 'service.get', 'foo']
            await s_cellauth.main(argv, outp)
            outp.expect(f'0 allow: service.get')
            outp.expect(f'1 deny: node.add')

            outp.clear()
            argv = [coreurl, 'modify', '--delrule', '1', 'foo']
            await s_cellauth.main(argv, outp)
            outp.expect(f'deleting rule index: 1')
            outp.expect(f'0 allow: service.get')
            self.false(outp.expect(f'deny: node.add', throw=False))

            outp.clear()
            argv = [coreurl, 'modify', '--delrule', '42', 'foo']
            await s_cellauth.main(argv, outp)
            outp.expect(f'deleting rule index: 42')
            outp.expect(f'rule index is out of range')

    async def test_cellauth_gates(self):

        async with self.getTestCore() as core:

            lurl = core.getLocalUrl()

            viewiden = core.view.iden
            layriden = core.view.layers[0].iden

            visi = await core.auth.addUser('visi')
            ninjas = await core.auth.addRole('ninjas')

            outp = self.getTestOutp()
            argv = [lurl, 'modify', '--addrule', 'node.add', '--object', layriden, 'visi']
            await s_cellauth.main(argv, outp)

            outp = self.getTestOutp()
            argv = [lurl, 'modify', '--admin', '--object', layriden, 'visi']
            await s_cellauth.main(argv, outp)

            outp = self.getTestOutp()
            argv = [lurl, 'modify', '--addrule', 'view.read', '--object', viewiden, 'ninjas']
            await s_cellauth.main(argv, outp)

            outp = self.getTestOutp()
            argv = [lurl, 'list', '--detail', 'ninjas']
            await s_cellauth.main(argv, outp)

            outp.expect(f'auth gate: {viewiden}')
            outp.expect('allow: view.read')

            outp = self.getTestOutp()
            argv = [lurl, 'list', '--detail', 'visi']
            await s_cellauth.main(argv, outp)

            outp.expect(f'auth gate: {layriden}')
            outp.expect('allow: node.add')

            outp.expect(f'auth gate: {viewiden}')
            outp.expect('allow: view.read')
